import ast
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

from boa3.analyser.astanalyser import IAstAnalyser
from boa3.exception import CompilerError
from boa3.model.builtin.builtin import Builtin
from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.callable import Callable
from boa3.model.importsymbol import Import
from boa3.model.method import Method
from boa3.model.module import Module
from boa3.model.operation.binary.binaryoperation import BinaryOperation
from boa3.model.operation.binaryop import BinaryOp
from boa3.model.operation.operation import IOperation
from boa3.model.operation.operator import Operator
from boa3.model.operation.unary.unaryoperation import UnaryOperation
from boa3.model.operation.unaryop import UnaryOp
from boa3.model.symbol import ISymbol
from boa3.model.type.collection.icollection import ICollectionType as Collection
from boa3.model.type.type import IType, Type
from boa3.model.variable import Variable


class TypeAnalyser(IAstAnalyser, ast.NodeVisitor):
    """
    This class is responsible for the type checking of the code

    The methods with the name starting with 'visit_' are implementations of methods from the :class:`NodeVisitor` class.
    These methods are used to walk through the Python abstract syntax tree.

    :ivar type_errors: a list with the found type errors. Empty by default.
    :ivar modules: a list with the analysed modules. Empty by default.
    :ivar symbols: a dictionary that maps the global symbols.
    :cvar _operators: a dictionary that maps each operator from Python ast to its equivalent Boa operator.
    """

    def __init__(self, analyser, symbol_table: Dict[str, ISymbol], log: bool = False):
        super().__init__(analyser.ast_tree, log=log)
        self.type_errors: List[Exception] = []
        self.modules: Dict[str, Module] = {}
        self.symbols: Dict[str, ISymbol] = symbol_table

        from boa3.builtin import NeoMetadata
        self._metadata: NeoMetadata = analyser.metadata
        self._current_method: Method = None

        self.visit(self._tree)

    _operators = {
        ast.Add: Operator.Plus,
        ast.Sub: Operator.Minus,
        ast.Mult: Operator.Mult,
        ast.Div: Operator.Div,
        ast.FloorDiv: Operator.IntDiv,
        ast.Mod: Operator.Mod,
        ast.Pow: Operator.Pow,
        ast.UAdd: Operator.Plus,
        ast.USub: Operator.Minus,
        ast.Eq: Operator.Eq,
        ast.NotEq: Operator.NotEq,
        ast.Lt: Operator.Lt,
        ast.LtE: Operator.LtE,
        ast.Gt: Operator.Gt,
        ast.GtE: Operator.GtE,
        ast.Is: Operator.Is,
        ast.IsNot: Operator.IsNot,
        ast.And: Operator.And,
        ast.Or: Operator.Or,
        ast.Not: Operator.Not,
        ast.BitAnd: Operator.BitAnd,
        ast.BitOr: Operator.BitOr,
        ast.BitXor: Operator.BitXor,
        ast.Invert: Operator.BitNot,
        ast.LShift: Operator.LeftShift,
        ast.RShift: Operator.RightShift
    }

    @property
    def _current_method_id(self) -> str:
        """
        Get the string identifier of the current method

        :return: The name identifier of the method. If the current method is None, returns None.
        :rtype: str or None
        """
        if self._current_method in self.symbols.values():
            index = list(self.symbols.values()).index(self._current_method)
            return list(self.symbols.keys())[index]

    @property
    def _modules_symbols(self) -> Dict[str, ISymbol]:
        """
        Gets all the symbols in the modules scopes.

        :return: Returns a dictionary that maps the modules symbols.
        """
        symbols = {}
        for module in self.modules.values():
            symbols.update(module.symbols)
        return symbols

    def get_symbol(self, symbol_id: str) -> Optional[ISymbol]:
        if not isinstance(symbol_id, str):
            return Variable(self.get_type(symbol_id))
        if self._current_method is not None and symbol_id in self._current_method.symbols:
            # the symbol exists in the local scope
            return self._current_method.symbols[symbol_id]
        elif symbol_id in self.modules:
            # the symbol exists in the modules scope
            return self.modules[symbol_id]
        else:
            return super().get_symbol(symbol_id)

    def visit_Module(self, module: ast.Module):
        """
        Visitor of the module node

        Performs the type checking in the body of the method

        :param module: the python ast module node
        """
        for stmt in module.body:
            self.visit(stmt)

    def visit_FunctionDef(self, function: ast.FunctionDef):
        """
        Visitor of the function node

        Performs the type checking in the body of the function

        :param function: the python ast function definition node
        """
        self.visit(function.args)
        method = self.symbols[function.name]
        from boa3.model.event import Event
        if isinstance(method, Method):
            self._current_method = method

            for stmt in function.body:
                self.visit(stmt)

            self._validate_return(function)

            if (len(function.body) > 0
                    and not isinstance(function.body[-1], ast.Return)
                    and method.return_type is Type.none):
                # include return None in void functions
                default_value: str = str(method.return_type.default_value)
                node: ast.AST = ast.parse(default_value).body[0].value
                function.body.append(
                    ast.Return(lineno=function.lineno, col_offset=function.col_offset, value=node)
                )
            self._current_method = None
        elif (isinstance(method, Event)         # events don't have return
                and function.returns is not None):
            return_type = self.get_type(function.returns)
            if return_type is not Type.none:
                self._log_error(
                    CompilerError.MismatchedTypes(line=function.lineno, col=function.col_offset,
                                                  expected_type_id=Type.none.identifier,
                                                  actual_type_id=return_type.identifier)
                )

    def visit_arguments(self, arguments: ast.arguments):
        """
        Verifies if each argument of a function has a type annotation

        :param arguments: the python ast function arguments node
        """
        for arg in arguments.args:
            self.visit(arg)

        # continue to walk through the tree
        self.generic_visit(arguments)

    def visit_arg(self, arg: ast.arg):
        """
        Verifies if the argument of a function has a type annotation

        :param arg: the python ast arg node
        """
        if arg.annotation is None:
            self._log_error(
                CompilerError.TypeHintMissing(arg.lineno, arg.col_offset, symbol_id=arg.arg)
            )

        # continue to walk through the tree
        self.generic_visit(arg)

    def visit_Return(self, ret: ast.Return):
        """
        Verifies if the return of the function is the same type as the return type annotation

        :param ret: the python ast return node
        """
        ret_value: Any = self.visit(ret.value) if ret.value is not None else None
        if ret.value is not None and self.get_type(ret.value) is not Type.none:
            # multiple returns are not allowed
            if isinstance(ret.value, ast.Tuple):
                self._log_error(
                    CompilerError.TooManyReturns(ret.lineno, ret.col_offset)
                )
                return
            # it is returning something, but there is no type hint for return
            elif self._current_method.return_type is Type.none:
                self._log_error(
                    CompilerError.TypeHintMissing(ret.lineno, ret.col_offset, symbol_id=self._current_method_id)
                )
                return
        self.validate_type_variable_assign(ret, ret_value, self._current_method)

        # continue to walk through the tree
        self.generic_visit(ret)

    def _validate_return(self, node: ast.AST):
        if (self._current_method.return_type is not Type.none
                and hasattr(node, 'body')
                and not self._has_return(node)):
            lst = list(self.symbols.items())
            keys, values = [key_value[0] for key_value in lst], [key_value[-1] for key_value in lst]
            self._log_error(
                CompilerError.MissingReturnStatement(
                    line=node.lineno, col=node.col_offset,
                    symbol_id=keys[values.index(self._current_method)]
                )
            )

    def _has_return(self, node: ast.AST) -> bool:
        if not hasattr(node, 'body'):
            return False

        has_else_stmt: bool = hasattr(node, 'orelse')
        has_body_return: bool = any(isinstance(stmt, (ast.Return, ast.Pass)) for stmt in node.body)
        if has_body_return and not has_else_stmt:
            # if any statement in the function body is a return, all flow has a return
            return True

        body = [stmt for stmt in node.body if hasattr(stmt, 'body')]
        body_has_inner_return: bool = len(body) > 0 and all(self._has_return(stmt) for stmt in body)
        if not has_body_return and not body_has_inner_return:
            # for and while nodes must to check if there is a return inside the else statement
            if not isinstance(node, (ast.For, ast.AsyncFor, ast.While)):
                return False

        if has_else_stmt:
            if any(isinstance(stmt, (ast.Return, ast.Pass)) for stmt in node.orelse):
                return True
            else:
                orelse = [stmt for stmt in node.orelse if hasattr(stmt, 'body')]
                return len(orelse) > 0 and all(self._has_return(stmt) for stmt in orelse)
        return body_has_inner_return

    def visit_Assign(self, assign: ast.Assign):
        """
        Verifies if it is a multiple assignments statement

        :param assign: the python ast variable assignment node
        """
        # multiple assignments
        if len(assign.targets) > 1:
            self._log_error(
                CompilerError.NotSupportedOperation(assign.lineno, assign.col_offset, 'Multiple variable assignments')
            )
        # multiple assignments with tuples
        elif isinstance(assign.targets[0], ast.Tuple):
            self._log_error(
                CompilerError.NotSupportedOperation(assign.lineno, assign.col_offset, 'Multiple variable assignments')
            )
        else:
            self.validate_type_variable_assign(assign.targets[0], assign.value)

        # continue to walk through the tree
        self.generic_visit(assign)

    def visit_AnnAssign(self, ann_assign: ast.AnnAssign):
        """
        Verifies if the assigned type is the same as the variable type

        :param ann_assign: the python ast variable annotated assignment node
        """
        # if value is None, it is a declaration
        if ann_assign.value is not None:
            self.validate_type_variable_assign(ann_assign.target, ann_assign.value)

    def visit_AugAssign(self, aug_assign: ast.AugAssign):
        """
        Verifies if the types of the target variable and the value are compatible with the operation

        If the operation is valid, changes de Python operator by the Boa operator in the syntax tree

        :param aug_assign: the python ast augmented assignment node
        """
        operation = self.validate_binary_operation(aug_assign, aug_assign.target, aug_assign.value)
        if operation is not None:
            # TODO: remove when other augmented assignment operations are implemented
            if operation.result not in [Type.int, Type.str]:
                raise NotImplementedError

            self.validate_type_variable_assign(aug_assign.target, operation)
            aug_assign.op = operation

    def validate_type_variable_assign(self, node: ast.AST, value: Any, target: Any = None) -> bool:
        value_type: IType = self.get_type(value)
        if isinstance(value, ast.AST):
            value = self.visit(value)

        if target is not None:
            target_type = self.get_type(target)
        elif not isinstance(node, ast.Name):
            target_type = self.get_type(node)
        else:
            var: ISymbol = self.get_symbol(node.id)
            if not isinstance(var, Variable):
                self._log_error(
                    CompilerError.UnresolvedReference(
                        node.lineno, node.col_offset,
                        symbol_id=node.id
                    ))
                return False
            if var.type is None:
                # it is an declaration with assignment and the value is neither literal nor another variable
                var.set_type(value_type)
            target_type = var.type

        if not target_type.is_type_of(value_type) and value != target_type.default_value:
            self._log_error(
                CompilerError.MismatchedTypes(
                    node.lineno, node.col_offset,
                    actual_type_id=value_type.identifier,
                    expected_type_id=target_type.identifier
                ))
            return False

        return True

    def visit_Subscript(self, subscript: ast.Subscript) -> IType:
        """
        Verifies if the subscribed value is a sequence

        :param subscript: the python ast subscript node
        :return: the type of the accessed value if it is valid. Type.none otherwise.
        """
        if isinstance(subscript.slice, ast.Index):
            return self.validate_get_or_set(subscript, subscript.slice)
        elif isinstance(subscript.slice, ast.Slice):
            return self.validate_slice(subscript, subscript.slice)
        return Type.none

    def validate_get_or_set(self, subscript: ast.Subscript, index_node: ast.Index) -> IType:
        """
        Verifies if the subscribed value is a sequence and if the index is valid to this sequence

        :param subscript: the python ast subscript node
        :param index_node: the subscript index
        :return: the type of the accessed value if it is valid. Type.none otherwise.
        """
        value = self.visit(subscript.value)
        index = self.visit(index_node)

        if isinstance(value, ast.Name):
            value = self.get_symbol(value.id)
        if isinstance(index, ast.Name):
            index = self.get_symbol(index.id)
        if not isinstance(index, tuple):
            index = (index,)

        # if it is a type hint, returns the outer type
        if isinstance(value, IType) and all(isinstance(i, IType) for i in index):
            return value

        symbol_type: IType = self.get_type(value)
        index_type: IType = self.get_type(index[0])

        # only sequence types can be subscribed
        if not isinstance(symbol_type, Collection):
            self._log_error(
                CompilerError.UnresolvedOperation(
                    subscript.lineno, subscript.col_offset,
                    type_id=symbol_type.identifier,
                    operation_id=Operator.Subscript)
            )
            return symbol_type
        # the sequence can't use the given type as index
        if not symbol_type.is_valid_key(index_type):
            self._log_error(
                CompilerError.MismatchedTypes(
                    subscript.lineno, subscript.col_offset,
                    actual_type_id=index_type.identifier,
                    expected_type_id=symbol_type.valid_key.identifier)
            )
        # it is setting a value in a sequence that doesn't allow reassign values
        elif isinstance(subscript.ctx, ast.Store) and not symbol_type.can_reassign_values:
            self._log_error(
                CompilerError.UnresolvedOperation(
                    subscript.lineno, subscript.col_offset,
                    type_id=symbol_type.identifier,
                    operation_id=Operator.Subscript)
            )
        return symbol_type.item_type

    def validate_slice(self, subscript: ast.Subscript, slice_node: ast.Slice) -> IType:
        """
        Verifies if the subscribed value is a sequence and if the slice is valid to this sequence

        :param subscript: the python ast subscript node
        :param slice_node: the subscript slice
        :return: the type of the accessed value if it is valid. Type.none otherwise.
        """
        value = self.visit(subscript.value)
        lower, upper, step = (self.get_type(value) for value in self.visit(slice_node))

        if step is not Type.none:
            # TODO: remove when slices with stride are implemented
            raise NotImplementedError
        # is not allowed to store into a slice
        if isinstance(subscript.ctx, ast.Store):
            self._log_error(
                CompilerError.NotSupportedOperation(
                    subscript.lineno, subscript.col_offset,
                    symbol_id=Operator.Subscript
                )
            )

        symbol_type: IType = self.get_type(value)
        # only collection types can be subscribed
        if not isinstance(symbol_type, Collection):
            self._log_error(
                CompilerError.UnresolvedOperation(
                    subscript.lineno, subscript.col_offset,
                    type_id=symbol_type.identifier,
                    operation_id=Operator.Subscript)
            )
            return Type.none

        lower = lower if lower is not Type.none else symbol_type.valid_key
        upper = upper if upper is not Type.none else symbol_type.valid_key

        # TODO: remove when slices of other sequence types are implemented
        if (not symbol_type.is_valid_key(lower)
            or not symbol_type.is_valid_key(upper)
            or (step is not Type.none and not symbol_type.is_valid_key(step))
            ):
            actual: Tuple[IType, ...] = (lower, upper) if step is Type.none else (lower, upper, step)
            self._log_error(
                CompilerError.MismatchedTypes(
                    subscript.lineno, subscript.col_offset,
                    expected_type_id=[symbol_type.valid_key.identifier for value in actual],
                    actual_type_id=[value.identifier for value in actual]
                )
            )
        else:
            return symbol_type

        return Type.none

    def visit_While(self, while_node: ast.While):
        """
        Verifies if the type of while test is valid

        :param while_node: the python ast while statement node
        """
        test = self.visit(while_node.test)
        test_type: IType = self.get_type(test)

        if test_type is not Type.bool:
            self._log_error(
                CompilerError.MismatchedTypes(
                    while_node.lineno, while_node.col_offset,
                    actual_type_id=test_type.identifier,
                    expected_type_id=Type.bool.identifier)
            )

        # continue to walk through the tree
        for stmt in while_node.body:
            self.visit(stmt)
        for stmt in while_node.orelse:
            self.visit(stmt)

    def visit_For(self, for_node: ast.For):
        """
        Verifies if the type of for iterator is valid

        :param for_node: the python ast for node
        """
        iterator = self.visit(for_node.iter)
        iterator_type: IType = self.get_type(iterator)

        if not isinstance(iterator_type, Collection):
            self._log_error(
                CompilerError.MismatchedTypes(
                    for_node.lineno, for_node.col_offset,
                    actual_type_id=iterator_type.identifier,
                    expected_type_id=Type.sequence.identifier)
            )

        # TODO: change when optimizing for loops
        elif self.get_type(for_node.target) != iterator_type.item_type:
            target_id = self.visit(for_node.target)
            if isinstance(target_id, ast.Name):
                target_id = target_id.id
            symbol = self.get_symbol(target_id)
            symbol.set_type(iterator_type.item_type)

        # continue to walk through the tree
        for stmt in for_node.body:
            self.visit(stmt)
        for stmt in for_node.orelse:
            self.visit(stmt)

    def visit_If(self, if_node: ast.If):
        """
        Verifies if the type of if test is valid

        :param if_node: the python ast if statement node
        """
        self.validate_if(if_node)
        # continue to walk through the tree
        for stmt in if_node.body:
            self.visit(stmt)
        for stmt in if_node.orelse:
            self.visit(stmt)

    def visit_IfExp(self, if_node: ast.IfExp):
        """
        Verifies if the type of if test is valid

        :param if_node: the python ast if expression node
        """
        self.validate_if(if_node)
        body = if_node.body
        orelse = if_node.orelse
        if_value = body[-1] if isinstance(body, list) and len(body) > 0 else body
        else_value = orelse[-1] if isinstance(orelse, list) and len(orelse) > 0 else orelse

        if_type: IType = self.get_type(if_value)
        else_type: IType = self.get_type(else_value)
        return Type.get_generic_type(if_type, else_type)

    def validate_if(self, if_node: ast.AST):
        """
        Verifies if the type of if test is valid

        :param if_node: the python ast if statement node
        :type if_node: ast.If or ast.IfExp
        """
        test = self.visit(if_node.test)
        test_type: IType = self.get_type(test)

        if test_type is not Type.bool:
            self._log_error(
                CompilerError.MismatchedTypes(
                    if_node.lineno, if_node.col_offset,
                    actual_type_id=test_type.identifier,
                    expected_type_id=Type.bool.identifier)
            )

    def visit_BinOp(self, bin_op: ast.BinOp) -> Optional[IType]:
        """
        Verifies if the types of the operands are valid to the operation

        If the operation is valid, changes de Python operator by the Boa operator in the syntax tree

        :param bin_op: the python ast binary operation node
        :return: the type of the result of the operation if the operation is valid. Otherwise, returns None
        :rtype: IType or None
        """
        operation = self.validate_binary_operation(bin_op, bin_op.left, bin_op.right)
        if operation is not None:
            bin_op.op = operation
            return operation.result

    def validate_binary_operation(self, node: ast.AST, left_op: ast.AST, right_op: ast.AST) -> Optional[IOperation]:
        """
        Validates a ast node that represents a binary operation

        :param node: ast node that represents a binary operation
        :param left_op: ast node that represents the left operand of the operation
        :param right_op: ast node that represents the right operand of the operation
        :return: the corresponding :class:`BinaryOperation` if is valid. None otherwise.
        """
        if not hasattr(node, 'op'):
            return

        operator: Operator = self.get_operator(node.op)
        l_operand = self.visit(left_op)
        r_operand = self.visit(right_op)

        if not isinstance(operator, Operator):
            # the operator is invalid or it was not implemented yet
            self._log_error(
                CompilerError.UnresolvedReference(node.lineno, node.col_offset, type(node.op).__name__)
            )

        try:
            operation: IOperation = self.get_bin_op(operator, r_operand, l_operand)
            if operation is None:
                self._log_error(
                    CompilerError.NotSupportedOperation(node.lineno, node.col_offset, operator)
                )
            elif not operation.is_supported:
                # TODO: concat and power not implemented yet
                # number float division is not supported by Neo VM
                self._log_error(
                    CompilerError.NotSupportedOperation(node.lineno, node.col_offset, operator)
                )
            else:
                return operation
        except CompilerError.MismatchedTypes as raised_error:
            raised_error.line = node.lineno
            raised_error.col = node.col_offset
            # raises the exception with the line/col info
            self._log_error(raised_error)

    def get_bin_op(self, operator: Operator, right: Any, left: Any) -> IOperation:
        """
        Returns the binary operation specified by the operator and the types of the operands

        :param operator: the operator
        :param right: right operand
        :param left: left operand

        :return: Returns the corresponding :class:`BinaryOperation` if the types are valid.
        :raise MismatchedTypes: raised if the types aren't valid for the operator
        """
        l_type: IType = self.get_type(left)
        r_type: IType = self.get_type(right)

        actual_types = (l_type.identifier, r_type.identifier)
        operation: IOperation = BinaryOp.validate_type(operator, l_type, r_type)

        if operation is not None:
            return operation
        else:
            expected_op: BinaryOperation = BinaryOp.get_operation_by_operator(operator, l_type)
            expected_types = (expected_op.left_type.identifier, expected_op.right_type.identifier)
            raise CompilerError.MismatchedTypes(0, 0, expected_types, actual_types)

    def visit_UnaryOp(self, un_op: ast.UnaryOp) -> Optional[IType]:
        """
        Verifies if the type of the operand is valid to the operation

        If the operation is valid, changes de Python operator by the Boa operator in the syntax tree

        :param un_op: the python ast unary operation node
        :return: the type of the result of the operation if the operation is valid. Otherwise, returns None
        :rtype: IType or None
        """
        operator: Operator = self.get_operator(un_op.op)
        operand = self.visit(un_op.operand)

        if not isinstance(operator, Operator):
            # the operator is invalid or it was not implemented yet
            self._log_error(
                CompilerError.UnresolvedReference(un_op.lineno, un_op.col_offset, type(un_op.op).__name__)
            )

        try:
            operation: UnaryOperation = self.get_un_op(operator, operand)
            if operation is None:
                self._log_error(
                    CompilerError.NotSupportedOperation(un_op.lineno, un_op.col_offset, operator)
                )
            elif not operation.is_supported:
                self._log_error(
                    CompilerError.NotSupportedOperation(un_op.lineno, un_op.col_offset, operator)
                )
            else:
                un_op.op = operation
                return operation.result
        except CompilerError.MismatchedTypes as raised_error:
            raised_error.line = un_op.lineno
            raised_error.col = un_op.col_offset
            # raises the exception with the line/col info
            self._log_error(raised_error)

    def get_un_op(self, operator: Operator, operand: Any) -> UnaryOperation:
        """
        Returns the binary operation specified by the operator and the types of the operands

        :param operator: the operator
        :param operand: the operand

        :return: Returns the corresponding :class:`UnaryOperation` if the types are valid.
        :raise MismatchedTypes: raised if the types aren't valid for the operator
        """
        op_type: IType = self.get_type(operand)

        actual_type: str = op_type.identifier
        operation: UnaryOperation = UnaryOp.validate_type(operator, op_type)

        if operation is not None:
            return operation
        else:
            expected_op: UnaryOperation = UnaryOp.get_operation_by_operator(operator)
            expected_type: str = expected_op.operand_type.identifier
            raise CompilerError.MismatchedTypes(0, 0, expected_type, actual_type)

    def visit_Compare(self, compare: ast.Compare) -> Optional[IType]:
        """
        Verifies if the types of the operands are valid to the compare operations

        If the operations are valid, changes de Python operator by the Boa operator in the syntax tree

        :param compare: the python ast compare operation node
        :return: the type of the result of the operation if the operation is valid. Otherwise, returns None
        :rtype: IType or None
        """
        if len(compare.comparators) != len(compare.ops):
            self._log_error(
                CompilerError.IncorrectNumberOfOperands(
                    compare.lineno,
                    compare.col_offset,
                    len(compare.ops),
                    len(compare.comparators) + 1  # num comparators + compare.left
                )
            )

        line = compare.lineno
        col = compare.col_offset
        try:
            return_type = None
            l_operand = self.visit_value(compare.left)
            for index, op in enumerate(compare.ops):
                operator: Operator = self.get_operator(op)
                r_operand = self.visit_value(compare.comparators[index])

                if not isinstance(operator, Operator):
                    # the operator is invalid or it was not implemented yet
                    self._log_error(
                        CompilerError.UnresolvedReference(line, col, type(op).__name__)
                    )

                operation: IOperation = self.get_bin_op(operator, r_operand, l_operand)
                if operation is None:
                    self._log_error(
                        CompilerError.NotSupportedOperation(line, col, operator)
                    )
                elif not operation.is_supported:
                    # TODO: is, is not and eq were not implemented yet
                    self._log_error(
                        CompilerError.NotSupportedOperation(line, col, operator)
                    )
                else:
                    compare.ops[index] = operation
                    return_type = operation.result

                line = compare.comparators[index].lineno
                col = compare.comparators[index].col_offset
                l_operand = r_operand

            return return_type
        except CompilerError.MismatchedTypes as raised_error:
            raised_error.line = line
            raised_error.col = col
            # raises the exception with the line/col info
            self._log_error(raised_error)

    def visit_BoolOp(self, bool_op: ast.BoolOp) -> Optional[IType]:
        """
        Verifies if the types of the operands are valid to the boolean operations

        If the operations are valid, changes de Python operator by the Boa operator in the syntax tree

        :param bool_op: the python ast boolean operation node
        :return: the type of the result of the operation if the operation is valid. Otherwise, returns None
        :rtype: IType or None
        """
        lineno: int = bool_op.lineno
        col_offset: int = bool_op.col_offset
        try:
            return_type: IType = None
            bool_operation: IOperation = None
            operator: Operator = self.get_operator(bool_op.op)

            if not isinstance(operator, Operator):
                # the operator is invalid or it was not implemented yet
                self._log_error(
                    CompilerError.UnresolvedReference(lineno, col_offset, type(operator).__name__)
                )

            l_operand = self.visit(bool_op.values[0])
            for index, operand in enumerate(bool_op.values[1:]):
                r_operand = self.visit(operand)

                operation: IOperation = self.get_bin_op(operator, r_operand, l_operand)
                if operation is None:
                    self._log_error(
                        CompilerError.NotSupportedOperation(lineno, col_offset, operator)
                    )
                elif bool_operation is None:
                    return_type = operation.result
                    bool_operation = operation

                lineno = operand.lineno
                col_offset = operand.col_offset
                l_operand = r_operand

            bool_op.op = bool_operation
            return return_type
        except CompilerError.MismatchedTypes as raised_error:
            raised_error.line = lineno
            raised_error.col = col_offset
            # raises the exception with the line/col info
            self._log_error(raised_error)

    def get_operator(self, node: Union[ast.operator, Operator, IOperation]) -> Optional[Operator]:
        """
        Gets the :class:`Operator` equivalent to given Python ast operator

        :param node: object with the operator data
        :type node: ast.operator or Operator or IOperation
        :return: the Boa operator equivalent to the node. None if it doesn't exit.
        :rtype: Operator or None
        """
        # the node has already been visited
        if isinstance(node, Operator):
            return node
        elif isinstance(node, IOperation):
            return node.operator

        node_type = type(node)
        if node_type in self._operators:
            return self._operators[node_type]
        else:
            return None

    def visit_Call(self, call: ast.Call):
        """
        Verifies if the number of arguments is correct

        :param call: the python ast function call node
        :return: the result type of the called function
        """
        if isinstance(call.func, ast.Name):
            callable_id: str = call.func.id
            callable_target = self.get_symbol(callable_id)
        else:
            arg0, callable_target, callable_id = self.visit(call.func)
            arg0_identifier = self.visit(arg0)
            if isinstance(arg0_identifier, ast.Name):
                arg0_identifier = arg0_identifier.id

            if (callable_target is not None
                    and (len(call.args) < 1 or call.args[0] != arg0)
                    and not isinstance(self.get_symbol(arg0_identifier), (IType, Import))):
                call.args.insert(0, arg0)
            if len(call.args) > 0 and isinstance(callable_target, IBuiltinMethod) and callable_target.has_self_argument:
                self_type: IType = self.get_type(call.args[0])
                caller = self.get_symbol(arg0_identifier)
                if isinstance(caller, IType) and not caller.is_type_of(self_type):
                    self_type = caller
                callable_target = callable_target.build(self_type)

        if not isinstance(callable_target, Callable):
            # verify if it is a builtin method with its name shadowed
            call_target = Builtin.get_symbol(callable_id)
            callable_target = call_target if call_target is not None else callable_target
        if not isinstance(callable_target, Callable):
            # the symbol doesn't exists or is not a function
            self._log_error(
                CompilerError.UnresolvedReference(call.func.lineno, call.func.col_offset, callable_id)
            )
        else:
            if callable_target is Builtin.NewEvent:
                return callable_target.return_type
            # TODO: change when kwargs is implemented
            if len(call.keywords) > 0:
                raise NotImplementedError

            len_call_args = len(call.args)
            callable_required_args = len(callable_target.args_without_default)
            if len_call_args > len(callable_target.args):
                unexpected_arg = call.args[len(callable_target.args)]
                self._log_error(
                    CompilerError.UnexpectedArgument(unexpected_arg.lineno, unexpected_arg.col_offset)
                )
            elif len_call_args < callable_required_args:
                missed_arg = list(callable_target.args)[len(call.args)]
                self._log_error(
                    CompilerError.UnfilledArgument(call.lineno, call.col_offset, missed_arg)
                )
            else:
                if callable_required_args <= len_call_args < len(callable_target.args):
                    included_args = len_call_args - callable_required_args
                    call.args.extend(callable_target.defaults[included_args:])
                args = [self.get_type(param) for param in call.args]
                if isinstance(callable_target, IBuiltinMethod):
                    # if the arguments are not generic, build the specified method
                    callable_target = callable_target.build(args)
                    if not callable_target.is_supported:
                        # TODO: implement bytearray constructor with non-bytes values
                        self._log_error(
                            CompilerError.NotSupportedOperation(call.lineno, call.col_offset, callable_id)
                        )
                        return callable_target.return_type

                for index, (arg_id, arg_value) in enumerate(callable_target.args.items()):
                    param = call.args[index]
                    param_type = self.get_type(param)
                    args.append(param_type)
                    if not arg_value.type.is_type_of(param_type):
                        self._log_error(
                            CompilerError.MismatchedTypes(
                                param.lineno, param.col_offset,
                                arg_value.type.identifier,
                                param_type.identifier
                            ))

            # if the arguments are not generic, include the specified method in the symbol table
            if (isinstance(callable_target, IBuiltinMethod)
                    and callable_target.identifier != callable_id
                    and callable_target.identifier not in self.symbols):
                self.symbols[callable_target.identifier] = callable_target
                call.func = ast.Name(lineno=call.func.lineno, col_offset=call.func.col_offset,
                                     ctx=ast.Load(), id=callable_target.identifier)
            if hasattr(callable_target, 'requires_storage') and callable_target.requires_storage:
                if not self._metadata.has_storage:
                    self._log_error(
                        CompilerError.MetadataInformationMissing(
                            line=call.func.lineno, col=call.func.col_offset,
                            symbol_id=callable_target.identifier,
                            metadata_attr_id='has_storage'
                        )
                    )
                else:
                    self._current_method.set_storage()
        return self.get_type(callable_target)

    def visit_value(self, node: ast.AST):
        result = self.visit(node)

        if isinstance(node, ast.Attribute):
            if isinstance(result, str):
                return self.get_symbol(result)
            else:
                origin, value, attribute = result
                if value is None and isinstance(origin, ast.Name):
                    value = self.get_symbol(origin.id)
                if value is not None:
                    return value

        return result

    def visit_Attribute(self, attribute: ast.Attribute) -> Union[str, Tuple[ast.AST, Optional[ISymbol], str]]:
        """
        Gets the attribute inside the ast node

        :param attribute: the python ast attribute node
        :return: returns the type of the value, the attribute symbol and its id if the attribute exists.
                 Otherwise, returns None
        """
        value: Optional[Union[str, ISymbol]] = self.get_symbol(attribute.value.id) \
            if isinstance(attribute.value, ast.Name) else self.visit(attribute.value)

        if value is None and isinstance(attribute.value, ast.Name):
            return '{0}.{1}'.format(attribute.value.id, attribute.attr)

        symbol = None
        if isinstance(value, str) and not isinstance(attribute.value, ast.Str):
            symbol = self.get_symbol(value)
            if symbol is None:
                return '{0}.{1}'.format(value, attribute.attr)
        if isinstance(value, ISymbol):
            symbol = value

        if hasattr(symbol, 'symbols') and attribute.attr in symbol.symbols:
            attr_symbol = symbol.symbols[attribute.attr]
        else:
            attr_symbol: Optional[ISymbol] = self.get_symbol(attribute.attr)

        return attribute.value, attr_symbol, attribute.attr

    def visit_Constant(self, constant: ast.Constant) -> Any:
        """
        Visitor of constant values node

        :param constant: the python ast constant value node
        :return: the value of the constant
        """
        if isinstance(constant, ast.Num) and not isinstance(constant.value, int):
            # only integer numbers are allowed
            self._log_error(
                CompilerError.InvalidType(constant.lineno, constant.col_offset, symbol_id=type(constant.value).__name__)
            )
        return constant.value

    def visit_Num(self, num: ast.Num) -> int:
        """
        Verifies if the number is an integer

        :param num: the python ast number node
        :return: returns the value of the number
        """
        if not isinstance(num.n, int):
            # only integer numbers are allowed
            self._log_error(
                CompilerError.InvalidType(num.lineno, num.col_offset, symbol_id=type(num.n).__name__)
            )
        return num.n

    def visit_Str(self, string: ast.Str) -> str:
        """
        Visitor of literal string node

        :param string: the python ast string node
        :return: the value of the string
        """
        return string.s

    def visit_Bytes(self, bts: ast.Bytes) -> bytes:
        """
        Visitor of literal bytes node

        :param bts: the python ast bytes node
        :return: the value of the bytes
        """
        return bts.s

    def visit_Tuple(self, tup_node: ast.Tuple) -> Tuple[Any, ...]:
        """
        Visitor of literal tuple node

        :param tup_node: the python ast tuple node
        :return: the value of the tuple
        """
        return tuple(self.get_type(value) for value in tup_node.elts)

    def visit_List(self, list_node: ast.List) -> List[Any]:
        """
        Visitor of literal list node

        :param list_node: the python ast list node
        :return: the value of the list
        """
        return [self.get_type(value) for value in list_node.elts]

    def visit_Dict(self, dict_node: ast.Dict) -> Dict[Any, Any]:
        """
        Visitor of literal dict node

        :param dict_node: the python ast dict node
        :return: a list with each key and value type
        """
        dictionary = {}
        size = min(len(dict_node.keys), len(dict_node.values))
        for index in range(size):
            key = self.get_type(dict_node.keys[index])
            value = self.get_type(dict_node.values[index])
            if key in dictionary and dictionary[key] != value:
                dictionary[key] = Type.get_generic_type(dictionary[key], value)
            else:
                dictionary[key] = value
        return dictionary

    def visit_NameConstant(self, constant: ast.NameConstant) -> Any:
        """
        Visitor of constant names node

        :param constant: the python ast name constant node
        :return: the value of the constant
        """
        return constant.value

    def visit_Name(self, name: ast.Name) -> ast.Name:
        """
        Visitor of a name node

        :param name:
        :return: the object with the name node information
        """
        return name

    def visit_Index(self, index: ast.Index) -> Any:
        """
        Visitor of an index node

        :param index:
        :return: the object with the index value information
        """
        index = self.visit(index.value)
        if isinstance(index, (Iterable, Tuple)):
            return tuple(index)
        return index

    def visit_Slice(self, slice_node: ast.Slice) -> Tuple[Any, Any, Any]:
        """
        Visitor of an slice node

        :param slice_node:
        :return: the object with the index value information
        """
        return slice_node.lower, slice_node.upper, slice_node.step

    def visit_Break(self, break_node: ast.Break):
        """
        :param break_node: the python ast break statement node
        """
        # TODO: remove when implement break statement
        raise NotImplementedError

    def visit_Continue(self, continue_node: ast.Continue):
        """
        :param continue_node: the python ast continue statement node
        """
        # TODO: remove when implement continue statement
        raise NotImplementedError



import ast
from typing import Optional

from boa3.model.expression import IExpression
from boa3.model.type.itype import IType


class Variable(IExpression):
    """
    A class used to represent a variable

    :ivar var_type: the type of the variable.
    """

    def __init__(self, var_type: Optional[IType], origin_node: Optional[ast.AST] = None):
        super().__init__(origin_node)
        self._var_type: Optional[IType] = var_type

    @property
    def shadowing_name(self) -> str:
        return 'variable'

    @property
    def type(self) -> IType:
        return self._var_type

    def __str__(self) -> str:
        return str(self.type)

    def set_type(self, var_type: IType):
        """
        Sets a type for the variable

        :param var_type:
        """
        self._var_type = var_type



from typing import Optional

from boa3 import constants
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.opcode.OpcodeInformation import OpcodeInformation


class OpcodeInfo:
    @classmethod
    def get_info(cls, opcode: Opcode) -> Optional[OpcodeInformation]:
        """
        Gets a binary operation given the operator.

        :param opcode: Neo VM opcode
        :return: The opcode info if it exists. None otherwise
        :rtype: OpcodeInformation or None
        """
        for id, op in vars(cls).items():
            if isinstance(op, OpcodeInformation) and op.opcode is opcode:
                return op

    # region Constants

    PUSHINT8 = OpcodeInformation(Opcode.PUSHINT8)
    PUSHINT16 = OpcodeInformation(Opcode.PUSHINT16)
    PUSHINT32 = OpcodeInformation(Opcode.PUSHINT32)
    PUSHINT64 = OpcodeInformation(Opcode.PUSHINT64)
    PUSHINT128 = OpcodeInformation(Opcode.PUSHINT128)
    PUSHINT256 = OpcodeInformation(Opcode.PUSHINT256)

    # Convert the next four bytes to an address, and push the address onto the stack.
    PUSHA = OpcodeInformation(Opcode.PUSHA, 4)
    # The item null is pushed onto the stack.
    PUSHNULL = OpcodeInformation(Opcode.PUSHNULL)
    # The next byte contains the number of bytes to be pushed onto the stack.
    PUSHDATA1 = OpcodeInformation(Opcode.PUSHDATA1, 1, constants.ONE_BYTE_MAX_VALUE)
    # The next two bytes contain the number of bytes to be pushed onto the stack.
    PUSHDATA2 = OpcodeInformation(Opcode.PUSHDATA2, 2, constants.TWO_BYTES_MAX_VALUE)
    # The next four bytes contain the number of bytes to be pushed onto the stack.
    PUSHDATA4 = OpcodeInformation(Opcode.PUSHDATA4, 4, constants.FOUR_BYTES_MAX_VALUE)

    # The number -1 is pushed onto the stack.
    PUSHM1 = OpcodeInformation(Opcode.PUSHM1)
    # The number 0 is pushed onto the stack.
    PUSH0 = OpcodeInformation(Opcode.PUSH0)
    # The number 1 is pushed onto the stack.
    PUSH1 = OpcodeInformation(Opcode.PUSH1)
    # The number 2 is pushed onto the stack.
    PUSH2 = OpcodeInformation(Opcode.PUSH2)
    # The number 3 is pushed onto the stack.
    PUSH3 = OpcodeInformation(Opcode.PUSH3)
    # The number 4 is pushed onto the stack.
    PUSH4 = OpcodeInformation(Opcode.PUSH4)
    # The number 5 is pushed onto the stack.
    PUSH5 = OpcodeInformation(Opcode.PUSH5)
    # The number 6 is pushed onto the stack.
    PUSH6 = OpcodeInformation(Opcode.PUSH6)
    # The number 7 is pushed onto the stack.
    PUSH7 = OpcodeInformation(Opcode.PUSH7)
    # The number 8 is pushed onto the stack.
    PUSH8 = OpcodeInformation(Opcode.PUSH8)
    # The number 9 is pushed onto the stack.
    PUSH9 = OpcodeInformation(Opcode.PUSH9)
    # The number 10 is pushed onto the stack.
    PUSH10 = OpcodeInformation(Opcode.PUSH10)
    # The number 11 is pushed onto the stack.
    PUSH11 = OpcodeInformation(Opcode.PUSH11)
    # The number 12 is pushed onto the stack.
    PUSH12 = OpcodeInformation(Opcode.PUSH12)
    # The number 13 is pushed onto the stack.
    PUSH13 = OpcodeInformation(Opcode.PUSH13)
    # The number 14 is pushed onto the stack.
    PUSH14 = OpcodeInformation(Opcode.PUSH14)
    # The number 15 is pushed onto the stack.
    PUSH15 = OpcodeInformation(Opcode.PUSH15)
    # The number 16 is pushed onto the stack.
    PUSH16 = OpcodeInformation(Opcode.PUSH16)

    # endregion

    # region Flow control

    # The NOP operation does nothing. It is intended to fill in space if opcodes are patched.
    NOP = OpcodeInformation(Opcode.NOP)
    # Unconditionally transfers control to a target instruction. The target instruction is represented as a 1-byte
    # signed offset from the beginning of the current instruction.
    JMP = OpcodeInformation(Opcode.JMP, 1)
    # Unconditionally transfers control to a target instruction. The target instruction is represented as a 4-bytes
    # signed offset from the beginning of the current instruction.
    JMP_L = OpcodeInformation(Opcode.JMP_L, 4)
    # Transfers control to a target instruction if the value is True, not null, or non-zero. The target instruction
    # is represented as a 1-byte signed offset from the beginning of the current instruction.
    JMPIF = OpcodeInformation(Opcode.JMPIF, 1, stack_items=1)
    # Transfers control to a target instruction if the value is True, not null, or non-zero. The target instruction
    # is represented as a 4-bytes signed offset from the beginning of the current instruction.
    JMPIF_L = OpcodeInformation(Opcode.JMPIF_L, 4, stack_items=1)
    # Transfers control to a target instruction if the value is False, a null reference,
    # or zero. The target instruction is represented as a 1-byte signed offset from the beginning of the current
    # instruction.
    JMPIFNOT = OpcodeInformation(Opcode.JMPIFNOT, 1, stack_items=1)
    # Transfers control to a target instruction if the value is False, a null reference,
    # or zero. The target instruction is represented as a 4-bytes signed offset from the beginning of the current
    # instruction.
    JMPIFNOT_L = OpcodeInformation(Opcode.JMPIFNOT_L, 4, stack_items=1)
    # Transfers control to a target instruction if two values are equal. The target instruction is represented as a
    # 1-byte signed offset from the beginning of the current instruction.
    JMPEQ = OpcodeInformation(Opcode.JMPEQ, 1, stack_items=2)
    # Transfers control to a target instruction if two values are equal. The target instruction is represented as a
    # 4-bytes signed offset from the beginning of the current instruction.
    JMPEQ_L = OpcodeInformation(Opcode.JMPEQ_L, 4, stack_items=2)
    # Transfers control to a target instruction when two values are not equal. The target instruction is represented
    # as a 1-byte signed offset from the beginning of the current instruction.
    JMPNE = OpcodeInformation(Opcode.JMPNE, 1, stack_items=2)
    # Transfers control to a target instruction when two values are not equal. The target instruction is represented
    # as a 4-bytes signed offset from the beginning of the current instruction.
    JMPNE_L = OpcodeInformation(Opcode.JMPNE_L, 4, stack_items=2)
    # Transfers control to a target instruction if the first value is greater than the second value. The target
    # instruction is represented as a 1-byte signed offset from the beginning of the current instruction.
    JMPGT = OpcodeInformation(Opcode.JMPGT, 1, stack_items=2)
    # Transfers control to a target instruction if the first value is greater than the second value. The target
    # instruction is represented as a 4-bytes signed offset from the beginning of the current instruction.
    JMPGT_L = OpcodeInformation(Opcode.JMPGT_L, 4, stack_items=2)
    # Transfers control to a target instruction if the first value is greater than or equal to the second value. The
    # target instruction is represented as a 1-byte signed offset from the beginning of the current instruction.
    JMPGE = OpcodeInformation(Opcode.JMPGE, 1, stack_items=2)
    # Transfers control to a target instruction if the first value is greater than or equal to the second value. The
    # target instruction is represented as a 4-bytes signed offset from the beginning of the current instruction.
    JMPGE_L = OpcodeInformation(Opcode.JMPGE_L, 4, stack_items=2)
    # Transfers control to a target instruction if the first value is less than the second value. The target
    # instruction is represented as a 1-byte signed offset from the beginning of the current instruction.
    JMPLT = OpcodeInformation(Opcode.JMPLT, 1, stack_items=2)
    # Transfers control to a target instruction if the first value is less than the second value. The target
    # instruction is represented as a 4-bytes signed offset from the beginning of the current instruction.
    JMPLT_L = OpcodeInformation(Opcode.JMPLT_L, 4, stack_items=2)
    # Transfers control to a target instruction if the first value is less than or equal to the second value. The
    # target instruction is represented as a 1-byte signed offset from the beginning of the current instruction.
    JMPLE = OpcodeInformation(Opcode.JMPLE, 1, stack_items=2)
    # Transfers control to a target instruction if the first value is less than or equal to the second value. The
    # target instruction is represented as a 4-bytes signed offset from the beginning of the current instruction.
    JMPLE_L = OpcodeInformation(Opcode.JMPLE_L, 4, stack_items=2)
    # Calls the function at the target address which is represented as a 1-byte signed offset from the beginning of
    # the current instruction.
    CALL = OpcodeInformation(Opcode.CALL, 1)
    # Calls the function at the target address which is represented as a 4-bytes signed offset from the beginning of
    # the current instruction.
    CALL_L = OpcodeInformation(Opcode.CALL_L, 4)
    # Pop the address of a function from the stack, and call the function.
    CALLA = OpcodeInformation(Opcode.CALLA)
    # It turns the vm state to FAULT immediately, and cannot be caught.
    ABORT = OpcodeInformation(Opcode.ABORT)
    # Pop the top value of the stack, if it false, then exit vm execution and set vm state to FAULT.
    ASSERT = OpcodeInformation(Opcode.ASSERT)
    # Pop the top value of the stack, and throw it.
    THROW = OpcodeInformation(Opcode.THROW)
    # TRY CatchOffset(sbyte) FinallyOffset(sbyte). If there's no catch body, set CatchOffset 0. If there's no finally
    # body, set FinallyOffset 0.
    TRY = OpcodeInformation(Opcode.TRY)
    # TRY_L CatchOffset(int) FinallyOffset(int). If there's no catch body, set CatchOffset 0. If there's no finally
    # body, set FinallyOffset 0.
    TRY_L = OpcodeInformation(Opcode.TRY_L)
    # Ensures that the appropriate surrounding finally blocks are executed. And then unconditionally transfers
    # control to the specific target instruction, represented as a 1-byte signed offset from the beginning of the
    # current instruction.
    ENDTRY = OpcodeInformation(Opcode.ENDTRY)
    # Ensures that the appropriate surrounding finally blocks are executed. And then unconditionally transfers
    # control to the specific target instruction, represented as a 4-byte signed offset from the beginning of the
    # current instruction.
    ENDTRY_L = OpcodeInformation(Opcode.ENDTRY_L)
    # End finally, If no exception happen or be catched, vm will jump to the target instruction of ENDTRY/ENDTRY_L.
    # Otherwise vm will rethrow the exception to upper layer.
    ENDFINALLY = OpcodeInformation(Opcode.ENDFINALLY)
    # Returns from the current method.
    RET = OpcodeInformation(Opcode.RET)
    # Calls to an interop service.
    SYSCALL = OpcodeInformation(Opcode.SYSCALL, min_data_len=4)

    # endregion

    # region Stack

    # Puts the number of stack items onto the stack.
    DEPTH = OpcodeInformation(Opcode.DEPTH)
    # Removes the top stack item.
    DROP = OpcodeInformation(Opcode.DROP)
    # Removes the second-to-top stack item.
    NIP = OpcodeInformation(Opcode.NIP)
    # The item n back in the main stack is removed.
    XDROP = OpcodeInformation(Opcode.XDROP, stack_items=1)
    # Clear the stack
    CLEAR = OpcodeInformation(Opcode.CLEAR)
    # Duplicates the top stack item.
    DUP = OpcodeInformation(Opcode.DUP)
    # Copies the second-to-top stack item to the top.
    OVER = OpcodeInformation(Opcode.OVER)
    # The item n back in the stack is copied to the top.
    PICK = OpcodeInformation(Opcode.PICK, stack_items=1)
    # The item at the top of the stack is copied and inserted before the second-to-top item.
    TUCK = OpcodeInformation(Opcode.TUCK)
    # The top two items on the stack are swapped.
    SWAP = OpcodeInformation(Opcode.SWAP)
    # The top three items on the stack are rotated to the left.
    ROT = OpcodeInformation(Opcode.ROT)
    # The item n back in the stack is moved to the top.
    ROLL = OpcodeInformation(Opcode.ROLL)
    # Reverse the order of the top 3 items on the stack.
    REVERSE3 = OpcodeInformation(Opcode.REVERSE3)
    # Reverse the order of the top 4 items on the stack.
    REVERSE4 = OpcodeInformation(Opcode.REVERSE4)
    # Pop the number N on the stack, and reverse the order of the top N items on the stack.
    REVERSEN = OpcodeInformation(Opcode.REVERSEN, stack_items=1)

    # endregion

    # region Slot

    # Initialize the static field list for the current execution context.
    INITSSLOT = OpcodeInformation(Opcode.INITSSLOT, 1)
    # Initialize the argument slot and the local variable list for the current execution context.
    INITSLOT = OpcodeInformation(Opcode.INITSLOT, 2)
    # Loads the static field at index 0 onto the evaluation stack.
    LDSFLD0 = OpcodeInformation(Opcode.LDSFLD0)
    # Loads the static field at index 1 onto the evaluation stack.
    LDSFLD1 = OpcodeInformation(Opcode.LDSFLD1)
    # Loads the static field at index 2 onto the evaluation stack.
    LDSFLD2 = OpcodeInformation(Opcode.LDSFLD2)
    # Loads the static field at index 3 onto the evaluation stack.
    LDSFLD3 = OpcodeInformation(Opcode.LDSFLD3)
    # Loads the static field at index 4 onto the evaluation stack.
    LDSFLD4 = OpcodeInformation(Opcode.LDSFLD4)
    # Loads the static field at index 5 onto the evaluation stack.
    LDSFLD5 = OpcodeInformation(Opcode.LDSFLD5)
    # Loads the static field at index 6 onto the evaluation stack.
    LDSFLD6 = OpcodeInformation(Opcode.LDSFLD6)
    # Loads the static field at a specified index onto the evaluation stack. The index is represented as a 1-byte
    # unsigned integer.
    LDSFLD = OpcodeInformation(Opcode.LDSFLD, 1)
    # Stores the value on top of the evaluation stack in the static field list at index 0.
    STSFLD0 = OpcodeInformation(Opcode.STSFLD0)
    # Stores the value on top of the evaluation stack in the static field list at index 1.
    STSFLD1 = OpcodeInformation(Opcode.STSFLD1)
    # Stores the value on top of the evaluation stack in the static field list at index 2.
    STSFLD2 = OpcodeInformation(Opcode.STSFLD2)
    # Stores the value on top of the evaluation stack in the static field list at index 3.
    STSFLD3 = OpcodeInformation(Opcode.STSFLD3)
    # Stores the value on top of the evaluation stack in the static field list at index 4.
    STSFLD4 = OpcodeInformation(Opcode.STSFLD4)
    # Stores the value on top of the evaluation stack in the static field list at index 5.
    STSFLD5 = OpcodeInformation(Opcode.STSFLD5)
    # Stores the value on top of the evaluation stack in the static field list at index 6.
    STSFLD6 = OpcodeInformation(Opcode.STSFLD6)
    # Stores the value on top of the evaluation stack in the static field list at a specified index. The index is
    # represented as a 1-byte unsigned integer.
    STSFLD = OpcodeInformation(Opcode.STSFLD, 1)
    # Loads the local variable at index 0 onto the evaluation stack.
    LDLOC0 = OpcodeInformation(Opcode.LDLOC0)
    # Loads the local variable at index 1 onto the evaluation stack.
    LDLOC1 = OpcodeInformation(Opcode.LDLOC1)
    # Loads the local variable at index 2 onto the evaluation stack.
    LDLOC2 = OpcodeInformation(Opcode.LDLOC2)
    # Loads the local variable at index 3 onto the evaluation stack.
    LDLOC3 = OpcodeInformation(Opcode.LDLOC3)
    # Loads the local variable at index 4 onto the evaluation stack.
    LDLOC4 = OpcodeInformation(Opcode.LDLOC4)
    # Loads the local variable at index 5 onto the evaluation stack.
    LDLOC5 = OpcodeInformation(Opcode.LDLOC5)
    # Loads the local variable at index 6 onto the evaluation stack.
    LDLOC6 = OpcodeInformation(Opcode.LDLOC6)
    # Loads the local variable at a specified index onto the evaluation stack. The index is represented as a 1-byte
    # unsigned integer.
    LDLOC = OpcodeInformation(Opcode.LDLOC, 1)
    # Stores the value on top of the evaluation stack in the local variable list at index 0.
    STLOC0 = OpcodeInformation(Opcode.STLOC0)
    # Stores the value on top of the evaluation stack in the local variable list at index 1.
    STLOC1 = OpcodeInformation(Opcode.STLOC1)
    # Stores the value on top of the evaluation stack in the local variable list at index 2.
    STLOC2 = OpcodeInformation(Opcode.STLOC2)
    # Stores the value on top of the evaluation stack in the local variable list at index 3.
    STLOC3 = OpcodeInformation(Opcode.STLOC3)
    # Stores the value on top of the evaluation stack in the local variable list at index 4.
    STLOC4 = OpcodeInformation(Opcode.STLOC4)
    # Stores the value on top of the evaluation stack in the local variable list at index 5.
    STLOC5 = OpcodeInformation(Opcode.STLOC5)
    # Stores the value on top of the evaluation stack in the local variable list at index 6.
    STLOC6 = OpcodeInformation(Opcode.STLOC6)
    # Stores the value on top of the evaluation stack in the local variable list at a specified index. The index is
    # represented as a 1-byte unsigned integer.
    STLOC = OpcodeInformation(Opcode.STLOC, 1)
    # Loads the argument at index 0 onto the evaluation stack.
    LDARG0 = OpcodeInformation(Opcode.LDARG0)
    # Loads the argument at index 1 onto the evaluation stack.
    LDARG1 = OpcodeInformation(Opcode.LDARG1)
    # Loads the argument at index 2 onto the evaluation stack.
    LDARG2 = OpcodeInformation(Opcode.LDARG2)
    # Loads the argument at index 3 onto the evaluation stack.
    LDARG3 = OpcodeInformation(Opcode.LDARG3)
    # Loads the argument at index 4 onto the evaluation stack.
    LDARG4 = OpcodeInformation(Opcode.LDARG4)
    # Loads the argument at index 5 onto the evaluation stack.
    LDARG5 = OpcodeInformation(Opcode.LDARG5)
    # Loads the argument at index 6 onto the evaluation stack.
    LDARG6 = OpcodeInformation(Opcode.LDARG6)
    # Loads the argument at a specified index onto the evaluation stack. The index is represented as a 1-byte
    # unsigned integer.
    LDARG = OpcodeInformation(Opcode.LDARG, 1)
    # Stores the value on top of the evaluation stack in the argument slot at index 0.
    STARG0 = OpcodeInformation(Opcode.STARG0)
    # Stores the value on top of the evaluation stack in the argument slot at index 1.
    STARG1 = OpcodeInformation(Opcode.STARG1)
    # Stores the value on top of the evaluation stack in the argument slot at index 2.
    STARG2 = OpcodeInformation(Opcode.STARG2)
    # Stores the value on top of the evaluation stack in the argument slot at index 3.
    STARG3 = OpcodeInformation(Opcode.STARG3)
    # Stores the value on top of the evaluation stack in the argument slot at index 4.
    STARG4 = OpcodeInformation(Opcode.STARG4)
    # Stores the value on top of the evaluation stack in the argument slot at index 5.
    STARG5 = OpcodeInformation(Opcode.STARG5)
    # Stores the value on top of the evaluation stack in the argument slot at index 6.
    STARG6 = OpcodeInformation(Opcode.STARG6)
    # Stores the value on top of the evaluation stack in the argument slot at a specified index. The index is
    # represented as a 1-byte unsigned integer.
    STARG = OpcodeInformation(Opcode.STARG, 1)

    # endregion

    # region Splice

    NEWBUFFER = OpcodeInformation(Opcode.NEWBUFFER)
    MEMCPY = OpcodeInformation(Opcode.MEMCPY)
    # Concatenates two strings.
    CAT = OpcodeInformation(Opcode.CAT)
    # Returns a section of a string.
    SUBSTR = OpcodeInformation(Opcode.SUBSTR)
    # Keeps only characters left of the specified point in a string.
    LEFT = OpcodeInformation(Opcode.LEFT)
    # Keeps only characters right of the specified point in a string.
    RIGHT = OpcodeInformation(Opcode.RIGHT)

    # endregion

    # region Bitwise logic

    # Flips all of the bits in the input.
    INVERT = OpcodeInformation(Opcode.INVERT)
    # Boolean and between each bit in the inputs.
    AND = OpcodeInformation(Opcode.AND)
    # Boolean or between each bit in the inputs.
    OR = OpcodeInformation(Opcode.OR)
    # Boolean exclusive or between each bit in the inputs.
    XOR = OpcodeInformation(Opcode.XOR)
    # Returns 1 if the inputs are exactly equal, 0 otherwise.
    EQUAL = OpcodeInformation(Opcode.EQUAL)
    # Returns 1 if the inputs are not equal, 0 otherwise.
    NOTEQUAL = OpcodeInformation(Opcode.NOTEQUAL)

    # endregion

    # region Arithmetic

    # Puts the sign of top stack item on top of the main stack. If value is negative, put -1; if positive,
    # put 1; if value is zero, put 0.
    SIGN = OpcodeInformation(Opcode.SIGN)
    # The input is made positive.
    ABS = OpcodeInformation(Opcode.ABS)
    # The sign of the input is flipped.
    NEGATE = OpcodeInformation(Opcode.NEGATE)
    # 1 is added to the input.
    INC = OpcodeInformation(Opcode.INC)
    # 1 is subtracted from the input.
    DEC = OpcodeInformation(Opcode.DEC)
    # a is added to b.
    ADD = OpcodeInformation(Opcode.ADD)
    # b is subtracted from a.
    SUB = OpcodeInformation(Opcode.SUB)
    # a is multiplied by b.
    MUL = OpcodeInformation(Opcode.MUL)
    # a is divided by b.
    DIV = OpcodeInformation(Opcode.DIV)
    # Returns the remainder after dividing a by b.
    MOD = OpcodeInformation(Opcode.MOD)
    # Shifts a left b bits, preserving sign.
    SHL = OpcodeInformation(Opcode.SHL)
    # Shifts a right b bits, preserving sign.
    SHR = OpcodeInformation(Opcode.SHR)
    # If the input is 0 or 1, it is flipped. Otherwise the output will be 0.
    NOT = OpcodeInformation(Opcode.NOT)
    # If both a and b are not 0, the output is 1. Otherwise 0.
    BOOLAND = OpcodeInformation(Opcode.BOOLAND)
    # If a or b is not 0, the output is 1. Otherwise 0.
    BOOLOR = OpcodeInformation(Opcode.BOOLOR)
    # Returns 0 if the input is 0. 1 otherwise.
    NZ = OpcodeInformation(Opcode.NZ)
    # Returns 1 if the numbers are equal, 0 otherwise.
    NUMEQUAL = OpcodeInformation(Opcode.NUMEQUAL)
    # Returns 1 if the numbers are not equal, 0 otherwise.
    NUMNOTEQUAL = OpcodeInformation(Opcode.NUMNOTEQUAL)
    # Returns 1 if a is less than b, 0 otherwise.
    LT = OpcodeInformation(Opcode.LT)
    # Returns 1 if a is less than or equal to b, 0 otherwise.
    LE = OpcodeInformation(Opcode.LE)
    # Returns 1 if a is greater than b, 0 otherwise.
    GT = OpcodeInformation(Opcode.GT)
    # Returns 1 if a is greater than or equal to b, 0 otherwise.
    GE = OpcodeInformation(Opcode.GE)
    # Returns the smaller of a and b.
    MIN = OpcodeInformation(Opcode.MIN)
    # Returns the larger of a and b.
    MAX = OpcodeInformation(Opcode.MAX)
    # Returns 1 if x is within the specified range (left-inclusive), 0 otherwise.
    WITHIN = OpcodeInformation(Opcode.WITHIN)

    # endregion

    # region Compound-type

    # A value n is taken from top of main stack. The next n items on main stack are removed, put inside n-sized array
    # and this array is put on top of the main stack.
    PACK = OpcodeInformation(Opcode.PACK)
    # An array is removed from top of the main stack. Its elements are put on top of the main stack (in reverse
    # order) and the array size is also put on main stack.
    UNPACK = OpcodeInformation(Opcode.UNPACK)
    # An empty array (with size 0) is put on top of the main stack.
    NEWARRAY0 = OpcodeInformation(Opcode.NEWARRAY0)
    # A value n is taken from top of main stack. A null-filled array with size n is put on top of the main stack.
    NEWARRAY = OpcodeInformation(Opcode.NEWARRAY)
    # A value n is taken from top of main stack. An array of type T with size n is put on top of the main stack.
    NEWARRAY_T = OpcodeInformation(Opcode.NEWARRAY_T, 1)
    # An empty struct (with size 0) is put on top of the main stack.
    NEWSTRUCT0 = OpcodeInformation(Opcode.NEWSTRUCT0)
    # A value n is taken from top of main stack. A zero-filled struct with size n is put on top of the main stack.
    NEWSTRUCT = OpcodeInformation(Opcode.NEWSTRUCT)
    # A Map is created and put on top of the main stack.
    NEWMAP = OpcodeInformation(Opcode.NEWMAP)
    # An array is removed from top of the main stack. Its size is put on top of the main stack.
    SIZE = OpcodeInformation(Opcode.SIZE)
    # An input index n (or key) and an array (or map) are removed from the top of the main stack. Puts True on top of
    # main stack if array[n] (or map[n]) exist, and False otherwise.
    HASKEY = OpcodeInformation(Opcode.HASKEY)
    # A map is taken from top of the main stack. The keys of this map are put on top of the main stack.
    KEYS = OpcodeInformation(Opcode.KEYS)
    # A map is taken from top of the main stack. The values of this map are put on top of the main stack.
    VALUES = OpcodeInformation(Opcode.VALUES)
    # An input index n (or key) and an array (or map) are taken from main stack. Element array[n] (or map[n]) is put
    # on top of the main stack.
    PICKITEM = OpcodeInformation(Opcode.PICKITEM)
    # The item on top of main stack is removed and appended to the second item on top of the main stack.
    APPEND = OpcodeInformation(Opcode.APPEND)
    # A value v, index n (or key) and an array (or map) are taken from main stack. Attribution array[n]=v
    # (or map[n]=v) is performed.
    SETITEM = OpcodeInformation(Opcode.SETITEM)
    # An array is removed from the top of the main stack and its elements are reversed.
    REVERSEITEMS = OpcodeInformation(Opcode.REVERSEITEMS)
    # An input index n (or key) and an array (or map) are removed from the top of the main stack. Element array[n]
    # (or map[n]) is removed.
    REMOVE = OpcodeInformation(Opcode.REMOVE)
    # Remove all the items from the compound-type.
    CLEARITEMS = OpcodeInformation(Opcode.CLEARITEMS)

    # endregion

    # region Types

    # Returns true if the input is null. Returns false otherwise.
    ISNULL = OpcodeInformation(Opcode.ISNULL)
    # Returns true if the top item is of the specified type.
    ISTYPE = OpcodeInformation(Opcode.ISTYPE, 1)
    # Converts the top item to the specified type.
    CONVERT = OpcodeInformation(Opcode.CONVERT, 1)

    # endregion



from boa3.boa3 import Boa3
from boa3.exception.CompilerError import MismatchedTypes
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3_test.tests.boa_test import BoaTest


class TestWhile(BoaTest):

    def test_while_constant_condition(self):
        jmpif_address = Integer(6).to_byte_array(min_length=1, signed=True)
        jmp_address = Integer(-5).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH0      # a = 0
            + Opcode.STLOC0
            + Opcode.JMP        # begin while
            + jmpif_address
                + Opcode.LDLOC0     # a = a + 2
                + Opcode.PUSH2
                + Opcode.ADD
                + Opcode.STLOC0
            + Opcode.PUSH0
            + Opcode.JMPIF      # end while False
            + jmp_address
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/while_test/ConstantCondition.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_while_variable_condition(self):
        jmpif_address = Integer(6).to_byte_array(min_length=1, signed=True)
        jmp_address = Integer(-5).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x01'
            + Opcode.PUSH0      # a = 0
            + Opcode.STLOC0
            + Opcode.JMP        # begin while
            + jmpif_address
                + Opcode.LDLOC0     # a = a + 2
                + Opcode.PUSH2
                + Opcode.ADD
                + Opcode.STLOC0
            + Opcode.LDARG0
            + Opcode.JMPIF      # end while arg0
            + jmp_address
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/while_test/VariableCondition.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_while_mismatched_type_condition(self):
        path = '%s/boa3_test/test_sc/while_test/MismatchedTypeCondition.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_while_no_condition(self):
        path = '%s/boa3_test/test_sc/while_test/NoCondition.py' % self.dirname

        with self.assertRaises(SyntaxError):
            output = Boa3.compile(path)

    def test_nested_while(self):
        outer_jmpif_address = Integer(19).to_byte_array(min_length=1, signed=True)
        outer_jmp_address = Integer(-18).to_byte_array(min_length=1, signed=True)

        inner_jmpif_address = Integer(6).to_byte_array(min_length=1, signed=True)
        inner_jmp_address = Integer(-5).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT
            + b'\x02'
            + b'\x02'
            + Opcode.PUSH0      # c = 0
            + Opcode.STLOC0
            + Opcode.LDLOC0     # d = c
            + Opcode.STLOC1
            + Opcode.JMP        # begin outer while
            + outer_jmpif_address
                + Opcode.LDLOC0     # c = c + 2
                + Opcode.PUSH2
                + Opcode.ADD
                + Opcode.STLOC0
                + Opcode.JMP        # begin inner while
                + inner_jmpif_address
                    + Opcode.LDLOC1     # d = d + 3
                    + Opcode.PUSH3
                    + Opcode.ADD
                    + Opcode.STLOC1
                + Opcode.LDARG1
                + Opcode.JMPIF      # end inner while arg1
                + inner_jmp_address
                + Opcode.LDLOC0     # c = c + d
                + Opcode.LDLOC1
                + Opcode.ADD
                + Opcode.STLOC0
            + Opcode.LDARG0
            + Opcode.JMPIF      # end outer while arg0
            + outer_jmp_address
            + Opcode.LDLOC0     # return c
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/while_test/NestedWhile.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_while_else(self):
        jmpif_address = Integer(6).to_byte_array(min_length=1, signed=True)
        jmp_address = Integer(-5).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH0      # a = 0
            + Opcode.STLOC0
            + Opcode.JMP        # begin while
            + jmpif_address
                + Opcode.LDLOC0     # a = a + 2
                + Opcode.PUSH2
                + Opcode.ADD
                + Opcode.STLOC0
            + Opcode.PUSH0
            + Opcode.JMPIF      # end while False
            + jmp_address
            + Opcode.LDLOC0     # else
            + Opcode.PUSH1          # a = a + 1
            + Opcode.ADD
            + Opcode.STLOC0
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/while_test/WhileElse.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_while_relational_condition(self):
        jmpif_address = Integer(10).to_byte_array(min_length=1, signed=True)
        jmp_address = Integer(-11).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT
            + b'\x02'
            + b'\x02'
            + Opcode.PUSH0      # a = 0
            + Opcode.STLOC0
            + Opcode.PUSH0      # b = 0
            + Opcode.STLOC1
            + Opcode.JMP        # begin while
            + jmpif_address
                + Opcode.LDLOC0     # a = a + 2
                + Opcode.PUSH2
                + Opcode.ADD
                + Opcode.STLOC0
                + Opcode.LDLOC1     # b = b + 1
                + Opcode.PUSH1
                + Opcode.ADD
                + Opcode.STLOC1
            + Opcode.LDLOC1
            + Opcode.PUSH10
            + Opcode.LT
            + Opcode.JMPIF      # end while b < 10
            + jmp_address
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/while_test/RelationalCondition.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_while_multiple_relational_condition(self):
        jmpif_address = Integer(10).to_byte_array(min_length=1, signed=True)
        jmp_address = Integer(-15).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT
            + b'\x02'
            + b'\x02'
            + Opcode.PUSH0      # a = 0
            + Opcode.STLOC0
            + Opcode.PUSH0      # b = 0
            + Opcode.STLOC1
            + Opcode.JMP        # begin while
            + jmpif_address
                + Opcode.LDLOC0     # a = a + 2
                + Opcode.PUSH2
                + Opcode.ADD
                + Opcode.STLOC0
                + Opcode.LDLOC1     # b = b + 1
                + Opcode.PUSH1
                + Opcode.ADD
                + Opcode.STLOC1
            + Opcode.LDLOC1
            + Opcode.PUSH10
            + Opcode.LT
            + Opcode.PUSH10
            + Opcode.LDARG1
            + Opcode.LT
            + Opcode.BOOLAND
            + Opcode.JMPIF      # end while b < 10 < arg1
            + jmp_address
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/while_test/MultipleRelationalCondition.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_boa2_while_test(self):
        path = '%s/boa3_test/test_sc/while_test/WhileBoa2Test.py' % self.dirname

        # TODO: change test when 'break' are implemented
        with self.assertRaises(NotImplementedError):
            output = Boa3.compile(path)

    def test_boa2_while_test1(self):
        jmpif_address = Integer(6).to_byte_array(min_length=1, signed=True)
        jmp_address = Integer(-7).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH3          # j = 3
            + Opcode.STLOC0
            + Opcode.JMP            # while
            + jmpif_address
                + Opcode.LDLOC0         # j = j + 1
                + Opcode.PUSH1
                + Opcode.ADD
                + Opcode.STLOC0
            + Opcode.LDLOC0         # while j < 6
            + Opcode.PUSH6
            + Opcode.LT
            + Opcode.JMPIF
            + jmp_address
            + Opcode.LDLOC0         # return j
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/while_test/WhileBoa2Test1.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_boa2_while_test2(self):
        path = '%s/boa3_test/test_sc/while_test/WhileBoa2Test2.py' % self.dirname

        # TODO: change test when 'break' are implemented
        with self.assertRaises(NotImplementedError):
            output = Boa3.compile(path)



from typing import List, Tuple

from boa3.model.operation.binary.binaryoperation import BinaryOperation
from boa3.model.operation.operator import Operator
from boa3.model.type.type import IType, Type
from boa3.neo.vm.opcode.Opcode import Opcode


class StrMultiplication(BinaryOperation):
    """
    A class used to represent a string concatenation operation

    :ivar operator: the operator of the operation. Inherited from :class:`IOperation`
    :ivar left: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar right: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar result: the result type of the operation.  Inherited from :class:`IOperation`
    """
    _valid_types: List[IType] = [Type.str]

    def __init__(self, left: IType = Type.str, right: IType = Type.int):
        self.operator: Operator = Operator.Mult
        super().__init__(left, right)

    def validate_type(self, *types: IType) -> bool:
        if len(types) != self.number_of_operands:
            return False
        left: IType = types[0]
        right: IType = types[1]

        return left in self._valid_types and right is Type.int

    def _get_result(self, left: IType, right: IType) -> IType:
        if self.validate_type(left, right):
            return left
        else:
            return Type.none

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.neo.vm.type.Integer import Integer
        return [
            (Opcode.PUSHDATA1, Integer(0).to_byte_array(min_length=1)),  # concatString = ''
            (Opcode.ROT, b''),
            (Opcode.ROT, b''),
            (Opcode.JMP, Integer(7).to_byte_array()),       # while argInt > 0
            (Opcode.REVERSE3, b''),
            (Opcode.OVER, b''),
            (Opcode.CAT, b''),                                  # concatString += argString
            (Opcode.REVERSE3, b''),
            (Opcode.DEC, b''),                                  # argInt -= 1
            (Opcode.DUP, b''),
            (Opcode.PUSH0, b''),
            (Opcode.JMPGT, Integer(-7).to_byte_array()),   # return concatString
            (Opcode.DROP, b''),
            (Opcode.DROP, b'')
        ]



import ast
from typing import Dict

from boa3.analyser.importanalyser import ImportAnalyser
from boa3.model.method import Method
from boa3.model.symbol import ISymbol
from boa3.model.type.itype import IType
from boa3.model.variable import Variable


class Import(ISymbol):
    """
    A class used to represent an imported package

    :ivar variables: a dictionary that maps each variable with its name. Empty by default.
    :ivar methods: a dictionary that maps each method with its name. Empty by default.
    :ivar types: a dictionary that maps each type with its name. Empty by default.
    """

    def __init__(self, origin: str, syntax_tree: ast.AST, import_analyser: ImportAnalyser,
                 imported_symbols: Dict[str, ISymbol] = None):
        if imported_symbols is None:
            symbols = import_analyser.symbols
        else:
            symbols = import_analyser.export_symbols(list(imported_symbols.keys()))

        self.variables = {var_id: var for var_id, var in symbols.items() if isinstance(var, Variable)}
        self.methods = {fun_id: fun for fun_id, fun in symbols.items() if isinstance(fun, Method)}
        self.types = {type_id: tpe for type_id, tpe in symbols.items() if isinstance(tpe, IType)}
        self.imports = {alias: imprt for alias, imprt in symbols.items() if isinstance(imprt, Import)}
        self._other_symbols = {alias: symbol for alias, symbol in symbols.items()
                               if not isinstance(symbol, (Variable, Method, IType, Import))}

        self._symbols_not_imported = {alias: symbol for alias, symbol in import_analyser.symbols.items()
                                      if alias not in symbols}

        self.origin: str = origin
        self.ast: ast.AST = syntax_tree

    @property
    def shadowing_name(self) -> str:
        return 'module'

    @property
    def symbols(self) -> Dict[str, ISymbol]:
        symbol = {}
        symbol.update(self.variables)
        symbol.update(self.methods)
        symbol.update(self.types)
        symbol.update(self._other_symbols)
        symbol.update(self.imports)
        return symbol

    @property
    def all_symbols(self) -> Dict[str, ISymbol]:
        symbol = self.symbols.copy()
        symbol.update(self._symbols_not_imported)
        return symbol



from boa3.boa3 import Boa3
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest


class TestMultipleExpressions(BoaTest):

    def test_multiple_arithmetic_expressions(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x03'
            + b'\x02'
            + Opcode.PUSH1      # d = 1
            + Opcode.STLOC0
            + Opcode.PUSH2      # e = 2
            + Opcode.STLOC1
            + Opcode.LDARG0     # c = a + b
            + Opcode.LDARG1
            + Opcode.ADD
            + Opcode.STLOC2
            + Opcode.LDLOC2     # return c
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/arithmetic_test/MultipleExpressionsInLine.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_multiple_relational_expressions(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x03'
            + b'\x02'
            + Opcode.LDARG0     # is_equal = a == b
            + Opcode.LDARG1
            + Opcode.NUMEQUAL
            + Opcode.STLOC0
            + Opcode.LDARG0     # is_greater = a > b
            + Opcode.LDARG1
            + Opcode.GT
            + Opcode.STLOC1
            + Opcode.LDARG0     # is_less = a < b
            + Opcode.LDARG1
            + Opcode.LT
            + Opcode.STLOC2
            + Opcode.LDLOC0     # return not is_equal
            + Opcode.NOT
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/relational_test/MultipleExpressionsInLine.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_multiple_logic_expressions(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x03'
            + b'\x03'
            + Opcode.LDARG0     # a1 = a and b
            + Opcode.LDARG1
            + Opcode.BOOLAND
            + Opcode.STLOC0
            + Opcode.LDARG1     # b1 = b and c
            + Opcode.LDARG2
            + Opcode.BOOLAND
            + Opcode.STLOC1
            + Opcode.LDARG0     # c1 = a or c
            + Opcode.LDARG2
            + Opcode.BOOLOR
            + Opcode.STLOC2
            + Opcode.LDLOC0     # return a1 and not b1 and c1
            + Opcode.LDLOC1
            + Opcode.NOT
            + Opcode.BOOLAND
            + Opcode.LDLOC2
            + Opcode.BOOLAND
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/logical_test/MultipleExpressionsInLine.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_multiple_tuple_expressions(self):
        a = String('a').to_bytes()
        b = String('b').to_bytes()
        c = String('c').to_bytes()
        d = String('d').to_bytes()

        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x03'
            + b'\x01'
            + Opcode.PUSHDATA1  # items2 = ('a', 'b', 'c', 'd')
            + Integer(len(d)).to_byte_array() + d
            + Opcode.PUSHDATA1
            + Integer(len(c)).to_byte_array() + c
            + Opcode.PUSHDATA1
            + Integer(len(b)).to_byte_array() + b
            + Opcode.PUSHDATA1
            + Integer(len(a)).to_byte_array() + a
            + Opcode.PUSH4
            + Opcode.PACK
            + Opcode.STLOC0     # items2 = array
            + Opcode.LDARG0     # value = items1[0]
            + Opcode.PUSH0
                + Opcode.DUP
                + Opcode.SIGN
                + Opcode.PUSHM1
                + Opcode.JMPNE
                + Integer(5).to_byte_array(min_length=1, signed=True)
                + Opcode.OVER
                + Opcode.SIZE
                + Opcode.ADD
            + Opcode.PICKITEM
            + Opcode.STLOC1
            + Opcode.LDLOC1     # count = value + len(items2)
            + Opcode.LDLOC0
            + Opcode.SIZE
            + Opcode.ADD
            + Opcode.STLOC2
            + Opcode.LDLOC2     # return count
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/tuple_test/MultipleExpressionsInLine.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_multiple_list_expressions(self):
        one = String('1').to_bytes()
        four = String('4').to_bytes()

        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x03'
            + b'\x01'
            + Opcode.PUSHDATA1  # items2 = [False, '1', 2, 3, '4']
            + Integer(len(four)).to_byte_array() + four
            + Opcode.PUSH3
            + Opcode.PUSH2
            + Opcode.PUSHDATA1
            + Integer(len(one)).to_byte_array() + one
            + Opcode.PUSH0
            + Opcode.PUSH5
            + Opcode.PACK
            + Opcode.STLOC0     # items2 = array
            + Opcode.LDARG0     # value = items1[0]
            + Opcode.PUSH0
                + Opcode.DUP
                + Opcode.SIGN
                + Opcode.PUSHM1
                + Opcode.JMPNE
                + Integer(5).to_byte_array(min_length=1, signed=True)
                + Opcode.OVER
                + Opcode.SIZE
                + Opcode.ADD
            + Opcode.PICKITEM
            + Opcode.STLOC1
            + Opcode.LDLOC1     # count = value + len(items2)
            + Opcode.LDLOC0
            + Opcode.SIZE
            + Opcode.ADD
            + Opcode.STLOC2
            + Opcode.LDLOC2     # return count
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/list_test/MultipleExpressionsInLine.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)



from boa3.boa3 import Boa3
from boa3.exception.CompilerError import MismatchedTypes
from boa3.model.type.type import Type
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest


class TestFor(BoaTest):

    def test_for_tuple_condition(self):
        jmpif_address = Integer(19).to_byte_array(min_length=1, signed=True)
        jmp_address = Integer(-22).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT
            + b'\x02'
            + b'\x00'
            + Opcode.PUSH0      # a = 0
            + Opcode.STLOC0
            + Opcode.PUSH15     # for_sequence = (3, 5, 15)
            + Opcode.PUSH5
            + Opcode.PUSH3
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.PUSH0      # for_index = 0
            + Opcode.JMP        # begin for
            + jmpif_address
                + Opcode.OVER           # x = for_sequence[for_index]
                + Opcode.OVER
                    + Opcode.DUP
                    + Opcode.SIGN
                    + Opcode.PUSHM1
                    + Opcode.JMPNE
                    + Integer(5).to_byte_array(min_length=1, signed=True)
                    + Opcode.OVER
                    + Opcode.SIZE
                    + Opcode.ADD
                + Opcode.PICKITEM
                + Opcode.STLOC1
                + Opcode.LDLOC0         # a = a + x
                + Opcode.LDLOC1
                + Opcode.ADD
                + Opcode.STLOC0
                + Opcode.INC            # for_index = for_index + 1
            + Opcode.DUP        # if for_index < len(for_sequence)
            + Opcode.PUSH2
            + Opcode.PICK
            + Opcode.SIZE
            + Opcode.LT
            + Opcode.JMPIF      # end for
            + jmp_address
            + Opcode.DROP
            + Opcode.DROP
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/for_test/TupleCondition.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_for_string_condition(self):
        sequence = String('3515').to_bytes(min_length=1)
        jmpif_address = Integer(24).to_byte_array(min_length=1, signed=True)
        jmp_address = Integer(-27).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT
            + b'\x03'
            + b'\x00'
            + Opcode.PUSH0      # a = 0
            + Opcode.STLOC0
            + Opcode.PUSHDATA1  # b = ''
            + Integer(0).to_byte_array(min_length=1)
            + Opcode.STLOC1
            + Opcode.PUSHDATA1  # for_sequence = '3515'
            + Integer(len(sequence)).to_byte_array(min_length=1)
            + sequence
            + Opcode.PUSH0      # for_index = 0
            + Opcode.JMP
                + jmpif_address
                + Opcode.OVER           # x = for_sequence[for_index]
                + Opcode.OVER
                    + Opcode.DUP
                    + Opcode.SIGN
                    + Opcode.PUSHM1
                    + Opcode.JMPNE
                    + Integer(5).to_byte_array(min_length=1, signed=True)
                    + Opcode.OVER
                    + Opcode.SIZE
                    + Opcode.ADD
                + Opcode.PUSH1
                + Opcode.SUBSTR
                + Opcode.CONVERT
                + Type.str.stack_item
                + Opcode.STLOC2
                + Opcode.LDLOC0         # a = a + 1
                + Opcode.PUSH1
                + Opcode.ADD
                + Opcode.STLOC0
                + Opcode.LDLOC2         # b = x
                + Opcode.STLOC1
                + Opcode.INC            # for_index = for_index + 1
            + Opcode.DUP        # if for_index < len(for_sequence)
            + Opcode.PUSH2
            + Opcode.PICK
            + Opcode.SIZE
            + Opcode.LT
            + Opcode.JMPIF      # end for
            + jmp_address
            + Opcode.DROP
            + Opcode.DROP
            + Opcode.LDLOC1     # return b
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/for_test/StringCondition.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_for_variable_condition(self):
        jmpif_address = Integer(19).to_byte_array(min_length=1, signed=True)
        jmp_address = Integer(-22).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT
            + b'\x03'
            + b'\x00'
            + Opcode.PUSH0      # a = 0
            + Opcode.STLOC0
            + Opcode.PUSH15     # sequence = (3, 5, 15)
            + Opcode.PUSH5
            + Opcode.PUSH3
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC1
            + Opcode.LDLOC1     # for_sequence = sequence
            + Opcode.PUSH0      # for_index = 0
            + Opcode.JMP
            + jmpif_address
                + Opcode.OVER         # x = for_sequence[for_index]
                + Opcode.OVER
                    + Opcode.DUP
                    + Opcode.SIGN
                    + Opcode.PUSHM1
                    + Opcode.JMPNE
                    + Integer(5).to_byte_array(min_length=1, signed=True)
                    + Opcode.OVER
                    + Opcode.SIZE
                    + Opcode.ADD
                + Opcode.PICKITEM
                + Opcode.STLOC2
                + Opcode.LDLOC0         # a = a + x
                + Opcode.LDLOC2
                + Opcode.ADD
                + Opcode.STLOC0
                + Opcode.INC            # for_index = for_index + 1
            + Opcode.DUP        # if for_index < len(for_sequence)
            + Opcode.PUSH2
            + Opcode.PICK
            + Opcode.SIZE
            + Opcode.LT
            + Opcode.JMPIF      # end for
            + jmp_address
            + Opcode.DROP
            + Opcode.DROP
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/for_test/VariableCondition.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_for_mismatched_type_condition(self):
        path = '%s/boa3_test/test_sc/for_test/MismatchedTypeCondition.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_for_no_condition(self):
        path = '%s/boa3_test/test_sc/for_test/NoCondition.py' % self.dirname

        with self.assertRaises(SyntaxError):
            output = Boa3.compile(path)

    def test_nested_for(self):
        outer_jmpif_address = Integer(47).to_byte_array(min_length=1, signed=True)
        outer_jmp_address = Integer(-50).to_byte_array(min_length=1, signed=True)

        inner_jmpif_address = Integer(21).to_byte_array(min_length=1, signed=True)
        inner_jmp_address = Integer(-24).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT
            + b'\x04'
            + b'\x00'
            + Opcode.PUSH0      # a = 0
            + Opcode.STLOC0
            + Opcode.PUSH15     # sequence = (3, 5, 15)
            + Opcode.PUSH5
            + Opcode.PUSH3
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC1
            + Opcode.LDLOC1     # outer_for_sequence = sequence
            + Opcode.PUSH0      # outer_for_index = 0
            + Opcode.JMP
            + outer_jmpif_address
                + Opcode.OVER           # x = outer_for_sequence[outer_for_index]
                + Opcode.OVER
                    + Opcode.DUP
                    + Opcode.SIGN
                    + Opcode.PUSHM1
                    + Opcode.JMPNE
                    + Integer(5).to_byte_array(min_length=1, signed=True)
                    + Opcode.OVER
                    + Opcode.SIZE
                    + Opcode.ADD
                + Opcode.PICKITEM
                + Opcode.STLOC2
                    + Opcode.LDLOC1     # inner_for_sequence = sequence
                    + Opcode.PUSH0      # inner_for_index = 0
                + Opcode.JMP
                + inner_jmpif_address
                    + Opcode.OVER         # y = inner_for_sequence[inner_for_index]
                    + Opcode.OVER
                        + Opcode.DUP
                        + Opcode.SIGN
                        + Opcode.PUSHM1
                        + Opcode.JMPNE
                        + Integer(5).to_byte_array(min_length=1, signed=True)
                        + Opcode.OVER
                        + Opcode.SIZE
                        + Opcode.ADD
                    + Opcode.PICKITEM
                    + Opcode.STLOC3
                    + Opcode.LDLOC0         # a = a + x * y
                    + Opcode.LDLOC2
                    + Opcode.LDLOC3
                    + Opcode.MUL
                    + Opcode.ADD
                    + Opcode.STLOC0
                    + Opcode.INC            # inner_for_index = inner_for_index + 1
                + Opcode.DUP        # if inner_for_index < len(inner_for_sequence)
                + Opcode.PUSH2
                + Opcode.PICK
                + Opcode.SIZE
                + Opcode.LT
                + Opcode.JMPIF      # end inner_for
                + inner_jmp_address
                + Opcode.DROP
                + Opcode.DROP
                + Opcode.INC     # outer_for_index = outer_for_index + 1
            + Opcode.DUP        # if outer_for_index < len(outer_for_sequence)
            + Opcode.PUSH2
            + Opcode.PICK
            + Opcode.SIZE
            + Opcode.LT
            + Opcode.JMPIF      # end outer_for
            + outer_jmp_address
            + Opcode.DROP
            + Opcode.DROP
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/for_test/NestedFor.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_for_else(self):
        jmpif_address = Integer(19).to_byte_array(min_length=1, signed=True)
        jmp_address = Integer(-22).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT
            + b'\x03'
            + b'\x00'
            + Opcode.PUSH0      # a = 0
            + Opcode.STLOC0
            + Opcode.PUSH15     # sequence = (3, 5, 15)
            + Opcode.PUSH5
            + Opcode.PUSH3
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC1
            + Opcode.LDLOC1     # for_sequence = sequence
            + Opcode.PUSH0      # for_index = 0
            + Opcode.JMP
            + jmpif_address
                + Opcode.OVER           # x = for_sequence[for_index]
                + Opcode.OVER
                    + Opcode.DUP
                    + Opcode.SIGN
                    + Opcode.PUSHM1
                    + Opcode.JMPNE
                    + Integer(5).to_byte_array(min_length=1, signed=True)
                    + Opcode.OVER
                    + Opcode.SIZE
                    + Opcode.ADD
                + Opcode.PICKITEM
                + Opcode.STLOC2
                + Opcode.LDLOC0         # a = a + x
                + Opcode.LDLOC2
                + Opcode.ADD
                + Opcode.STLOC0
                + Opcode.INC            # for_index = for_index + 1
            + Opcode.DUP        # if for_index < len(for_sequence)
            + Opcode.PUSH2
            + Opcode.PICK
            + Opcode.SIZE
            + Opcode.LT
            + Opcode.JMPIF      # end for
            + jmp_address
            + Opcode.DROP
            + Opcode.DROP
            + Opcode.LDLOC0     # a = a + 1
            + Opcode.PUSH1
            + Opcode.ADD
            + Opcode.STLOC0
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/for_test/ForElse.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)



import ast
from abc import ABC
from typing import Dict, List, Tuple

from boa3.model.builtin.method.builtinevent import IBuiltinEvent
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class InteropEvent(IBuiltinEvent, ABC):

    def __init__(self, identifier: str, sys_call: str,
                 args: Dict[str, Variable] = None, defaults: List[ast.AST] = None):
        self._sys_call: str = sys_call
        super().__init__(identifier, args, defaults)

    @property
    def interop_method_hash(self) -> bytes:
        return self._method_hash(self._sys_call)

    def _method_hash(self, method_name: str) -> bytes:
        from boa3.constants import SIZE_OF_INT32
        from boa3.neo import cryptography
        from boa3.neo.vm.type.String import String

        return cryptography.sha256(String(method_name).to_bytes())[:SIZE_OF_INT32]

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        return [(Opcode.SYSCALL, self.interop_method_hash)]



import ast

from boa3.analyser.analyser import Analyser
from boa3.analyser.typeanalyser import TypeAnalyser
from boa3.model.type.collection.sequence.mutable.listtype import ListType
from boa3.model.type.collection.sequence.tupletype import TupleType
from boa3.model.type.type import Type
from boa3_test.tests.boa_test import BoaTest


class TestTypes(BoaTest):

    def test_small_integer_constant(self):
        input = 42
        node = ast.parse(str(input)).body[0].value
        expected_output = Type.int

        typeanalyser = TypeAnalyser(Analyser(node), {})
        output = typeanalyser.get_type(input)

        self.assertEqual(expected_output, output)

    def test_negative_integer_constant(self):
        input = -10
        node = ast.parse(str(input)).body[0].value
        expected_output = Type.int

        typeanalyser = TypeAnalyser(Analyser(node), {})
        output = typeanalyser.get_type(input)

        self.assertEqual(expected_output, output)

    def test_boolean_constant(self):
        input = True
        node = ast.parse(str(input)).body[0].value
        expected_output = Type.bool

        typeanalyser = TypeAnalyser(Analyser(node), {})
        output = typeanalyser.get_type(input)

        self.assertEqual(expected_output, output)

    def test_string_constant(self):
        input = 'unit_test'
        node = ast.parse(str(input)).body[0].value
        expected_output = Type.str

        typeanalyser = TypeAnalyser(Analyser(node), {})
        output = typeanalyser.get_type(input)

        self.assertEqual(expected_output, output)

    def test_integer_tuple_constant(self):
        input = (1, 2, 3)
        node = ast.parse(str(input)).body[0].value
        expected_output = TupleType(Type.int)

        typeanalyser = TypeAnalyser(Analyser(node), {})
        output = typeanalyser.get_type(input)

        self.assertEqual(expected_output, output)

    def test_boolean_tuple_constant(self):
        input = (True, False)
        node = ast.parse(str(input)).body[0].value
        expected_output = TupleType(Type.bool)

        typeanalyser = TypeAnalyser(Analyser(node), {})
        output = typeanalyser.get_type(input)

        self.assertEqual(expected_output, output)

    def test_string_tuple_constant(self):
        input = (1, 2, 3)
        node = ast.parse(str(input)).body[0].value
        expected_output = TupleType(Type.int)

        typeanalyser = TypeAnalyser(Analyser(node), {})
        output = typeanalyser.get_type(input)

        self.assertEqual(expected_output, output)

    def test_any_tuple_constant(self):
        input = (1, '2', False)
        node = ast.parse(str(input)).body[0].value
        expected_output = TupleType(Type.any)

        typeanalyser = TypeAnalyser(Analyser(node), {})
        output = typeanalyser.get_type(input)

        self.assertEqual(expected_output, output)

    def test_integer_list_constant(self):
        input = [1, 2, 3]
        node = ast.parse(str(input)).body[0].value
        expected_output = ListType(Type.int)

        typeanalyser = TypeAnalyser(Analyser(node), {})
        output = typeanalyser.get_type(input)

        self.assertEqual(expected_output, output)

    def test_boolean_list_constant(self):
        input = [True, False]
        node = ast.parse(str(input)).body[0].value
        expected_output = ListType(Type.bool)

        typeanalyser = TypeAnalyser(Analyser(node), {})
        output = typeanalyser.get_type(input)

        self.assertEqual(expected_output, output)

    def test_string_list_constant(self):
        input = [1, 2, 3]
        node = ast.parse(str(input)).body[0].value
        expected_output = ListType(Type.int)

        typeanalyser = TypeAnalyser(Analyser(node), {})
        output = typeanalyser.get_type(input)

        self.assertEqual(expected_output, output)

    def test_any_list_constant(self):
        input = [1, '2', False]
        node = ast.parse(str(input)).body[0].value
        expected_output = ListType(Type.any)
        
        typeanalyser = TypeAnalyser(Analyser(node), {})
        output = typeanalyser.get_type(input)

        self.assertEqual(expected_output, output)

    def test_sequence_any_is_type_of_str(self):
        sequence_type = Type.sequence
        str_type = Type.str
        self.assertTrue(sequence_type.is_type_of(str_type))

    def test_sequence_any_is_type_of_tuple_any(self):
        sequence_type = Type.sequence
        tuple_type = Type.tuple
        self.assertTrue(sequence_type.is_type_of(tuple_type))

    def test_sequence_int_is_type_of_tuple_any(self):
        sequence_type = Type.sequence.build_collection(Type.int)
        tuple_type = Type.tuple
        self.assertFalse(sequence_type.is_type_of(tuple_type))

    def test_sequence_any_is_type_of_tuple_int(self):
        sequence_type = Type.sequence
        tuple_type = Type.tuple.build_collection(Type.int)
        self.assertTrue(sequence_type.is_type_of(tuple_type))

    def test_sequence_any_is_type_of_list_any(self):
        sequence_type = Type.sequence
        list_type = Type.list
        self.assertTrue(sequence_type.is_type_of(list_type))

    def test_sequence_int_is_type_of_list_any(self):
        sequence_type = Type.sequence.build_collection(Type.int)
        list_type = Type.list
        self.assertFalse(sequence_type.is_type_of(list_type))

    def test_sequence_any_is_type_of_list_int(self):
        sequence_type = Type.sequence
        list_type = Type.list.build_collection(Type.int)
        self.assertTrue(sequence_type.is_type_of(list_type))

    def test_tuple_any_is_type_of_sequence(self):
        sequence_type = Type.sequence
        tuple_type = Type.tuple
        self.assertFalse(tuple_type.is_type_of(sequence_type))

    def test_tuple_any_is_type_of_tuple_int(self):
        tuple_type = Type.tuple
        tuple_int_type = Type.tuple.build_collection(Type.int)
        self.assertTrue(tuple_type.is_type_of(tuple_int_type))

    def test_tuple_int_is_type_of_tuple_any(self):
        tuple_type = Type.tuple.build_collection(Type.int)
        tuple_any_type = Type.tuple
        self.assertFalse(tuple_type.is_type_of(tuple_any_type))

    def test_list_any_is_type_of_sequence(self):
        list_type = Type.list
        sequence_type = Type.sequence
        self.assertFalse(list_type.is_type_of(sequence_type))

    def test_list_any_is_type_of_list_int(self):
        list_type = Type.list
        list_int_type = Type.list.build_collection(Type.int)
        self.assertTrue(list_type.is_type_of(list_int_type))

    def test_list_int_is_type_of_list_any(self):
        list_type = Type.list.build_collection(Type.int)
        list_any_type = Type.list
        self.assertFalse(list_type.is_type_of(list_any_type))

    def test_str_any_is_type_of_sequence(self):
        sequence_type = Type.sequence
        str_type = Type.str
        self.assertFalse(str_type.is_type_of(sequence_type))

    def test_str_any_is_type_of_sequence_str(self):
        sequence_type = Type.sequence.build_collection(Type.str)
        str_type = Type.str
        self.assertFalse(str_type.is_type_of(sequence_type))



from boa3.constants import ENCODING


class String(str):
    def to_bytes(self, min_length: int = 0) -> bytes:
        """
        Converts an string value to a immutable byte array

        :param min_length: Minimum output length.
        :return: Return an array of bytes representing an string.
        """

        bts = bytes(self, ENCODING)
        return bts.ljust(min_length)

    @classmethod
    def from_bytes(cls, bts: bytes) -> str:
        return bts.decode(ENCODING)



import ast
from typing import Dict, List, Optional, Tuple

from boa3.compiler.codegenerator import CodeGenerator
from boa3.compiler.vmcodemapping import VMCodeMapping
from boa3.model.builtin.builtin import Builtin
from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.method import Method
from boa3.model.operation.binary.binaryoperation import BinaryOperation
from boa3.model.operation.binaryop import BinaryOp
from boa3.model.operation.operation import IOperation
from boa3.model.operation.unary.unaryoperation import UnaryOperation
from boa3.model.symbol import ISymbol
from boa3.model.type.type import IType, Type
from boa3.model.variable import Variable


class VisitorCodeGenerator(ast.NodeVisitor):
    """
    This class is responsible for walk through the ast.

    The methods with the name starting with 'visit_' are implementations of methods from the :class:`NodeVisitor` class.
    These methods are used to walk through the Python abstract syntax tree.

    :ivar generator:
    """

    def __init__(self, generator: CodeGenerator):
        self.generator = generator
        self.current_method: Optional[Method] = None

    @property
    def symbols(self) -> Dict[str, ISymbol]:
        return self.generator.symbol_table

    def include_instruction(self, node: ast.AST, address: int):
        if self.current_method is not None and address in VMCodeMapping.instance().code_map:
            bytecode = VMCodeMapping.instance().code_map[address]
            from boa3.model.debuginstruction import DebugInstruction
            self.current_method.include_instruction(DebugInstruction.build(node, bytecode))

    def visit_to_map(self, node: ast.AST, generate: bool = False):
        address: int = VMCodeMapping.instance().bytecode_size
        if generate:
            value = self.visit_to_generate(node)
        else:
            value = self.visit(node)

        if not hasattr(node, 'test'):
            # control flow nodes must map each of their instructions
            self.include_instruction(node, address)
        return value

    def visit_to_generate(self, node):
        """
        Visitor to generate the nodes that the primary visitor is used to retrieve value

        :param node: an ast node
        """
        if isinstance(node, ast.AST):
            result = self.visit(node)
            # the default return of the name visitor is the name string
            if isinstance(result, str):
                # TODO: validate function calls
                self.generator.convert_load_symbol(result)
            return result
        else:
            return self.generator.convert_literal(node)

    def visit_Module(self, module: ast.Module):
        """
        Visitor of the module node

        Fills module symbol table

        :param module:
        """
        global_stmts = [node for node in module.body if not isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef))]
        function_stmts = module.body[len(global_stmts):]

        for stmt in function_stmts:
            self.visit(stmt)

        if self.generator.initialize_static_fields():
            for stmt in global_stmts:
                self.visit(stmt)

            self.generator.end_initialize()

    def visit_ImportFrom(self, import_from: ast.ImportFrom):
        """
        Includes methods and variables from other modules into the current scope

        :param import_from:
        """
        self._import_static_fields(import_from.names)

    def visit_Import(self, import_node: ast.Import):
        """
        Includes methods and variables from other modules into the current scope

        :param import_node:
        """
        self._import_static_fields(import_node.names)

    def _import_static_fields(self, names: List[ast.alias]):
        """
        Visits the imported nodes that aren't function definitions to initialize static fields

        :param names: list of imported alias
        """
        for alias in names:
            name = alias.asname if alias.asname is not None else alias.name
            if name in self.symbols:
                if hasattr(self.symbols[name], 'ast') and self.symbols[name].ast is not None:
                    ast_root = self.symbols[name].ast
                elif hasattr(self.symbols[name], 'origin') and self.symbols[name].origin is not None:
                    ast_root = self.symbols[name].origin
                else:
                    continue

                if isinstance(ast_root, ast.Module):
                    body = ast_root.body
                else:
                    body = [ast_root]

                for node in body:
                    if not isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
                        self.visit(node)

    def visit_FunctionDef(self, function: ast.FunctionDef):
        """
        Visitor of the function definition node

        Generates the Neo VM code for the function

        :param function: the python ast function definition node
        """
        method = self.symbols[function.name]
        if isinstance(method, Method):
            self.current_method = method
            self.generator.convert_begin_method(method)

            for stmt in function.body:
                self.visit_to_map(stmt)

            self.generator.convert_end_method()
            self.current_method = None

    def visit_arguments(self, arguments: ast.arguments) -> Dict[str, Variable]:
        """
        Visitor of the function arguments node

        :param arguments: the python ast function arguments node
        :return: a dictionary that maps each argument to its identifier
        """
        args: Dict[str, Variable] = {}

        for arg in arguments.args:
            var_id, var = self.visit_arg(arg)   # type:str, Variable
            args[var_id] = var
        return args

    def visit_arg(self, arg: ast.arg) -> Tuple[str, Variable]:
        """
        Visitor of a function argument node

        :param arg: the python ast arg node
        :return: a tuple with the identifier and the argument
        """
        var_id = arg.arg
        var_type = self.visit(arg.annotation)

        return var_id, Variable(var_type)

    def visit_Return(self, ret: ast.Return):
        """
        Visitor of a function return node

        :param ret: the python ast return node
        """
        if ret.value is not None:
            self.visit_to_generate(ret.value)
            self.generator.insert_return()

    def store_variable(self, var_id: str, value: ast.AST, index: ast.AST = None):
        # if the value is None, it is a variable declaration
        if value is not None:
            if index is None:
                # if index is None, then it is a variable assignment
                self.visit_to_generate(value)
                self.generator.convert_store_variable(var_id)
            else:
                # if not, it is an array assignment
                self.generator.convert_load_symbol(var_id)
                self.visit_to_generate(index)
                value_address = self.visit_to_generate(value)
                self.generator.convert_set_item(value_address)

    def visit_AnnAssign(self, ann_assign: ast.AnnAssign):
        """
        Visitor of an annotated assignment node

        :param ann_assign: the python ast variable assignment node
        """
        var_id = self.visit(ann_assign.target)
        self.store_variable(var_id, ann_assign.value)

    def visit_Assign(self, assign: ast.Assign):
        """
        Visitor of an assignment node

        :param assign: the python ast variable assignment node
        """
        var_index = None
        var_id = self.visit(assign.targets[0])

        # if it is a tuple, then it is an array assignment
        if isinstance(var_id, tuple):
            var_index = var_id[1]
            var_id: str = var_id[0]

        self.store_variable(var_id, assign.value, var_index)

    def visit_AugAssign(self, aug_assign: ast.AugAssign):
        """
        Visitor of an augmented assignment node

        :param aug_assign: the python ast augmented assignment node
        """
        var_id = self.visit(aug_assign.target)
        self.generator.convert_load_symbol(var_id)
        self.visit_to_generate(aug_assign.value)
        self.generator.convert_operation(aug_assign.op)
        self.generator.convert_store_variable(var_id)

    def visit_Subscript(self, subscript: ast.Subscript):
        """
        Visitor of a subscript node

        :param subscript: the python ast subscript node
        """
        if isinstance(subscript.slice, ast.Index):
            return self.visit_Subscript_Index(subscript)
        elif isinstance(subscript.slice, ast.Slice):
            return self.visit_Subscript_Slice(subscript)

    def visit_Subscript_Index(self, subscript: ast.Subscript):
        """
        Visitor of a subscript node with index

        :param subscript: the python ast subscript node
        """
        if isinstance(subscript.ctx, ast.Load):
            # get item
            self.visit_to_generate(subscript.value)
            self.visit_to_generate(subscript.slice.value)
            self.generator.convert_get_item()
        else:
            # set item
            var_id = self.visit(subscript.value)
            return var_id, subscript.slice.value

    def visit_Subscript_Slice(self, subscript: ast.Subscript):
        """
        Visitor of a subscript node with slice

        :param subscript: the python ast subscript node
        """
        lower_omitted = subscript.slice.lower is None
        upper_omitted = subscript.slice.upper is None

        self.visit_to_generate(subscript.value)
        # if both are explicit
        if not lower_omitted and not upper_omitted:
            addresses = [VMCodeMapping.instance().bytecode_size]
            self.visit_to_generate(subscript.slice.lower)

            # length of slice
            addresses.append(VMCodeMapping.instance().bytecode_size)
            self.visit_to_generate(subscript.slice.upper)

            self.generator.convert_get_sub_array(addresses)
        # only one of them is omitted
        elif lower_omitted != upper_omitted:
            # end position is omitted
            if lower_omitted:
                self.visit_to_generate(subscript.slice.upper)
                self.generator.convert_get_array_beginning()
            # start position is omitted
            else:
                self.generator.duplicate_stack_top_item()
                # length of slice
                self.generator.convert_builtin_method_call(Builtin.Len)
                self.visit_to_generate(subscript.slice.lower)
                self.generator.convert_get_array_ending()
        else:
            self.generator.convert_copy()

    def _convert_unary_operation(self, operand, op):
        self.visit_to_generate(operand)
        self.generator.convert_operation(op)

    def _convert_binary_operation(self, left, right, op):
        self.visit_to_generate(left)
        self.visit_to_generate(right)
        self.generator.convert_operation(op)

    def visit_BinOp(self, bin_op: ast.BinOp):
        """
        Visitor of a binary operation node

        :param bin_op: the python ast binary operation node
        """
        if isinstance(bin_op.op, BinaryOperation):
            self._convert_binary_operation(bin_op.left, bin_op.right, bin_op.op)

    def visit_UnaryOp(self, un_op: ast.UnaryOp):
        """
        Visitor of a binary operation node

        :param un_op: the python ast binary operation node
        """
        if isinstance(un_op.op, UnaryOperation):
            self._convert_unary_operation(un_op.operand, un_op.op)

    def visit_Compare(self, compare: ast.Compare):
        """
        Visitor of a compare operation node

        :param compare: the python ast compare operation node
        """
        converted: bool = False
        left = compare.left
        for index, op in enumerate(compare.ops):
            right = compare.comparators[index]
            if isinstance(op, IOperation):
                if isinstance(op, BinaryOperation):
                    self._convert_binary_operation(left, right, op)
                else:
                    operand = left
                    if isinstance(operand, ast.NameConstant) and operand.value is None:
                        operand = right
                    self._convert_unary_operation(operand, op)
                # if it's more than two comparators, must include AND between the operations
                if not converted:
                    converted = True
                else:
                    self.generator.convert_operation(BinaryOp.And)
            left = right

    def visit_BoolOp(self, bool_op: ast.BoolOp):
        """
        Visitor of a compare operation node

        :param bool_op: the python ast boolean operation node
        """
        if isinstance(bool_op.op, BinaryOperation):
            left = bool_op.values[0]
            self.visit_to_generate(left)
            for index, right in enumerate(bool_op.values[1:]):
                self.visit_to_generate(right)
                self.generator.convert_operation(bool_op.op)

    def visit_While(self, while_node: ast.While):
        """
        Visitor of a while statement node

        :param while_node: the python ast while statement node
        """
        start_addr: int = self.generator.convert_begin_while()
        for stmt in while_node.body:
            self.visit_to_map(stmt, generate=True)

        test_address: int = VMCodeMapping.instance().bytecode_size
        self.visit_to_map(while_node.test, generate=True)
        self.generator.convert_end_while(start_addr, test_address)

        for stmt in while_node.orelse:
            self.visit_to_map(stmt, generate=True)

    def visit_For(self, for_node: ast.For):
        """
        Visitor of for statement node

        :param for_node: the python ast for node
        """
        self.visit_to_generate(for_node.iter)
        start_address = self.generator.convert_begin_for()

        if isinstance(for_node.target, tuple):
            for target in for_node.target:
                var_id = self.visit_to_map(target)
                self.generator.convert_store_variable(var_id)
        else:
            var_id = self.visit(for_node.target)
            self.generator.convert_store_variable(var_id)

        for stmt in for_node.body:
            self.visit_to_map(stmt, generate=True)

        # TODO: remove when optimizing for generation
        if self.current_method is not None:
            self.current_method.remove_instruction(for_node.lineno, for_node.col_offset)

        condition_address = self.generator.convert_end_for(start_address)
        self.include_instruction(for_node, condition_address)

        for stmt in for_node.orelse:
            self.visit_to_map(stmt, generate=True)

    def visit_If(self, if_node: ast.If):
        """
        Visitor of if statement node

        :param if_node: the python ast if statement node
        """
        self.visit_to_map(if_node.test, generate=True)

        start_addr: int = self.generator.convert_begin_if()
        for stmt in if_node.body:
            self.visit_to_map(stmt, generate=True)

        if len(if_node.orelse) > 0:
            start_addr = self.generator.convert_begin_else(start_addr)
            for stmt in if_node.orelse:
                self.visit_to_map(stmt, generate=True)

        self.generator.convert_end_if(start_addr)

    def visit_IfExp(self, if_node: ast.IfExp):
        """
        Visitor of if expression node

        :param if_node: the python ast if statement node
        """
        self.visit_to_map(if_node.test, generate=True)

        start_addr: int = self.generator.convert_begin_if()
        self.visit_to_map(if_node.body, generate=True)

        start_addr = self.generator.convert_begin_else(start_addr)
        self.visit_to_map(if_node.orelse, generate=True)

        self.generator.convert_end_if(start_addr)

    def visit_Assert(self, assert_node: ast.Assert):
        """
        Visitor of the assert node

        :param assert_node: the python ast assert node
        """
        self.visit_to_generate(assert_node.test)
        self.generator.convert_assert()

    def visit_Call(self, call: ast.Call):
        """
        Visitor of a function call node

        :param call: the python ast function call node
        """
        # the parameters are included into the stack in the reversed order
        function_id = self.visit(call.func)
        symbol = self.generator.get_symbol(function_id)
        args_addresses: List[int] = []

        if isinstance(symbol, IBuiltinMethod) and symbol.push_self_first():
            args = call.args[1:]
            args_addresses.append(
                VMCodeMapping.instance().bytecode_size
            )
            self.visit_to_generate(call.args[0])
        else:
            args = call.args

        for arg in reversed(args):
            args_addresses.append(
                VMCodeMapping.instance().bytecode_size
            )
            self.visit_to_generate(arg)
        self.generator.convert_load_symbol(function_id, args_addresses)

    def visit_Name(self, name: ast.Name) -> str:
        """
        Visitor of a name node

        :param name: the python ast name identifier node
        :return: the identifier of the name
        """
        return name.id

    def visit_Attribute(self, attribute: ast.Attribute) -> str:
        """
        Visitor of a attribute node

        :param attribute: the python ast attribute node
        :return: the identifier of the attribute
        """
        if self.generator.get_symbol(attribute.attr) is not Type.none:
            return attribute.attr

        value = attribute.value
        if isinstance(value, ast.Attribute):
            value = self.visit(value)
        if isinstance(value, (ast.Name, str)):
            value_id = value.id if isinstance(value, ast.Name) else value
            return '{0}.{1}'.format(value_id, attribute.attr)
        return attribute.attr

    def visit_Constant(self, constant: ast.Constant):
        """
        Visitor of constant values node

        :param constant: the python ast constant value node
        """
        return self.generator.convert_literal(constant.value)

    def visit_NameConstant(self, constant: ast.NameConstant):
        """
        Visitor of constant names node

        :param constant: the python ast name constant node
        """
        return self.generator.convert_literal(constant.value)

    def visit_Num(self, num: ast.Num):
        """
        Visitor of literal number node

        :param num: the python ast number node
        """
        return self.generator.convert_literal(num.n)

    def visit_Str(self, string: ast.Str):
        """
        Visitor of literal string node

        :param string: the python ast string node
        """
        return self.generator.convert_literal(string.s)

    def visit_Bytes(self, bts: ast.Bytes):
        """
        Visitor of literal bytes node

        :param bts: the python ast bytes node
        """
        return self.generator.convert_literal(bts.s)

    def visit_Tuple(self, tup_node: ast.Tuple):
        """
        Visitor of literal tuple node

        :param tup_node: the python ast tuple node
        """
        self._create_array(tup_node.elts, Type.tuple)

    def visit_List(self, list_node: ast.List):
        """
        Visitor of literal list node

        :param list_node: the python ast list node
        """
        self._create_array(list_node.elts, Type.list)

    def visit_Dict(self, dict_node: ast.Dict):
        """
        Visitor of literal dict node

        :param dict_node: the python ast dict node
        """
        length = min(len(dict_node.keys), len(dict_node.values))
        self.generator.convert_new_map(Type.dict)
        for key_value in range(length):
            self.generator.duplicate_stack_top_item()
            self.visit_to_generate(dict_node.keys[key_value])
            value_address = self.visit_to_generate(dict_node.values[key_value])
            self.generator.convert_set_item(value_address)

    def _create_array(self, values: List[ast.AST], array_type: IType):
        """
        Creates a new array from a literal sequence

        :param values: list of values of the new array items
        """
        length = len(values)
        if length > 0:
            for value in reversed(values):
                self.visit_to_generate(value)
        self.generator.convert_new_array(length, array_type)



from typing import Dict, List

from boa3.neo.vm.VMCode import VMCode
from boa3.neo.vm.opcode.OpcodeInformation import OpcodeInformation


class VMCodeMapping:
    """
    This class is responsible for managing the Neo VM instruction during the bytecode generation.
    """
    _instance = None  # type:VMCodeMapping

    @classmethod
    def instance(cls):
        """
        :return: the singleton instance
        """
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def __init__(self):
        self._codes: Dict[int, VMCode] = {}

    @classmethod
    def reset(cls):
        """
        Resets the map to the first state
        """
        if cls._instance is not None:
            cls._instance._codes.clear()

    @property
    def codes(self) -> List[VMCode]:
        """
        Gets a list with the included vm codes

        :return: a list of vm codes ordered by its address in the bytecode
        """
        return list(self.code_map.values())

    @property
    def code_map(self) -> Dict[int, VMCode]:
        """
        Gets a dictionary that maps each vm code with its address.

        :return: a dictionary that maps each instruction with its address. The keys are ordered by the address.
        """
        return {key: self._codes[key] for key in sorted(self._codes)}

    def bytecode(self) -> bytes:
        """
        Gets the bytecode of the translated code

        :return: the generated bytecode
        """
        self._update_larger_codes()

        bytecode = bytearray()
        for code in self.codes:
            bytecode += code.opcode
            if code.data is not None:
                bytecode += code.data
        return bytes(bytecode)

    @property
    def bytecode_size(self) -> int:
        if len(self._codes) < 1:
            return 0

        last_key = list(self.code_map)[-1]
        return last_key + self._codes[last_key].size

    def insert_code(self, vm_code: VMCode):
        if vm_code not in self._codes.values():
            self._codes[self.bytecode_size] = vm_code

    def get_start_address(self, vm_code: VMCode) -> int:
        """
        Gets the vm code's first byte address

        :param vm_code: the instruction to get the address
        :return: the vm code's address if it's in the map. Otherwise, return's zero.
        """
        if vm_code not in self._codes.values():
            return 0
        return next(key for key, value in self._codes.items() if value == vm_code)

    def get_end_address(self, vm_code: VMCode) -> int:
        """
        Gets the vm code's last byte address

        :param vm_code: the instruction to get the address
        :return: the vm code's last address if it's in the map. Otherwise, return's zero.
        """
        if vm_code not in self._codes.values():
            return 0
        return self.get_start_address(vm_code) + vm_code.size - 1  # start + size returns next opcode address

    def update_vm_code(self, vm_code: VMCode, opcode: OpcodeInformation, data: bytes = bytes()):
        """
        Updates the information from an inserted code

        :param vm_code: code to be updated
        :param opcode: updated opcode information
        :param data: updated opcode data
        """
        code_size = vm_code.size
        vm_code._info = opcode
        vm_code._data = data
        if vm_code.size != code_size:
            self._update_addresses(self.get_start_address(vm_code))

    def _update_addresses(self, start_address: int = 0):
        """
        Updates the instruction map's keys when a opcode is changed

        :param start_address: the address from the changed opcode
        """
        new_address = -1
        last_code: VMCode = None
        updated_codes: Dict[int, VMCode] = {}

        for address, code in list(self.code_map.items()):
            if address >= start_address:
                if new_address < 0:
                    new_address = self.get_start_address(last_code) if last_code is not None else 0

                new_address += last_code.size
                if new_address != address:
                    updated_codes[new_address] = self._codes.pop(address)
            last_code = code
        self._codes.update(updated_codes)

    def move_to_end(self, first_code_address: int, last_code_address: int):
        """
        Moves a set of instructions to the end of the current bytecode

        :param first_code_address: first instruction start address
        :param last_code_address: last instruction end address
        """
        if last_code_address < first_code_address:
            return

        moved_codes: List[VMCode] = []
        for address in list(self.code_map):
            if first_code_address <= address <= last_code_address:
                moved_codes.append(self._codes.pop(address))
            elif address > last_code_address:
                break

        self._update_addresses(first_code_address)

        for code in moved_codes:
            self.insert_code(code)
        self._update_targets()

    def _update_targets(self):
        from boa3.neo.vm.type.Integer import Integer
        for address, code in self.code_map.items():
            if code.opcode.has_target() and code.target is None:
                relative = Integer.from_bytes(code.data)
                absolute = address + relative
                if absolute in self.code_map:
                    code.set_target(self.code_map[absolute])

    def _update_larger_codes(self):
        """
        Checks if each instruction data fits in its opcode maximum size and updates the opcode from those that don't
        """
        # gets a list with all instructions which its opcode has a larger equivalent, ordered by its address
        instr_with_small_codes = [code for code in self._codes.values() if code.opcode.has_larger_opcode()]
        instr_with_small_codes.sort(key=lambda code: self.get_start_address(code), reverse=True)

        from boa3.neo.vm.opcode.OpcodeInfo import OpcodeInfo
        # total_len is initialized with zero because the loop must run at least once
        total_len = 0
        current_size = self.bytecode_size

        # if any instruction is updated, the following instruction addresses and the total size will change as well
        # with the change, previous instruction data may have overflowed the opcode maximum value
        # to make sure, it must check the opcodes that haven't changed again
        while total_len != current_size:
            total_len = current_size

            # verifies each instruction data length
            for code in instr_with_small_codes.copy():  # it's a copy because the list may change during the iteration
                if len(code.raw_data) > code.info.max_data_len:
                    # gets the shortest opcode equivalent that fits the instruction data
                    info = OpcodeInfo.get_info(code.opcode.get_larger_opcode())
                    while len(code.raw_data) > info.max_data_len and info.opcode.has_larger_opcode():
                        info = OpcodeInfo.get_info(code.opcode.get_larger_opcode())

                    self.update_vm_code(code, info)
                    if info.opcode == info.opcode.get_larger_opcode():
                        # if it's the largest equivalent, it won't be updated anymore
                        instr_with_small_codes.remove(code)
            current_size = self.bytecode_size



import ast
import importlib.util
import os
from typing import Dict, List

from boa3.analyser.astanalyser import IAstAnalyser
from boa3.model.builtin.builtin import Builtin
from boa3.model.symbol import ISymbol
from boa3.model.type.type import Type


class ImportAnalyser(IAstAnalyser):

    def __init__(self, import_target: str):
        self.can_be_imported: bool = False
        self._import_identifier: str = import_target

        module_origin: str = importlib.util.find_spec(import_target).origin
        path: List[str] = module_origin.split(os.sep)
        self.path: str = module_origin.replace(os.sep, '/')

        super().__init__(ast.Module(body=[]), path[-1])
        if import_target == 'typing':
            self.symbols.update(
                {symbol_id: symbol for symbol_id, symbol in self._get_types_from_typing_lib().items()
                 if symbol_id not in Type.builtin_types()
                 })
            self.can_be_imported = True

        else:
            import re

            inside_python_folder = any(re.search(r'python(\d\.?)*', folder.lower()) for folder in path)
            updated_tree = None

            if 'boa3' in path and '/'.join(path[path.index('boa3'):]).startswith('boa3/builtin'):
                pkg_start_index = path.index('builtin') + 1
                if path[pkg_start_index] == path[-1]:
                    self.symbols = self._get_boa3_builtin_symbols()
                else:
                    pkg = import_target.split('.')
                    pkg = pkg[pkg.index('builtin') + 1:]
                    self.symbols = self._get_boa3_builtin_package(pkg)
                self.can_be_imported = True

            elif not (inside_python_folder and 'lib' in path):
                # TODO: only user modules and typing lib imports are implemented
                try:
                    from boa3.analyser.analyser import Analyser
                    analyser = Analyser.analyse(module_origin)

                    # include only imported symbols
                    if analyser.is_analysed:
                        self.symbols.update(
                            {symbol_id: symbol for symbol_id, symbol in analyser.symbol_table.items()
                             if symbol_id not in Type.all_types()
                             })
                    updated_tree = analyser.ast_tree
                    self.can_be_imported = analyser.is_analysed
                except FileNotFoundError:
                    self.can_be_imported = False

                if updated_tree is not None:
                    self._tree = updated_tree

    @property
    def tree(self) -> ast.AST:
        return self._tree

    def export_symbols(self, identifiers: List[str] = None) -> Dict[str, ISymbol]:
        """
        Gets a dictionary that maps each exported symbol with its identifier

        :param identifiers: list of identifiers of the imported symbols
        :return:
        """
        if identifiers is None:
            identifiers = list([name for name, symbol in self.symbols.items() if symbol is not None])
        if not self.can_be_imported or not isinstance(identifiers, (list, str)):
            return {}

        if isinstance(identifiers, str):
            identifiers = [identifiers]

        symbols = {symbol_id: symbol for symbol_id, symbol in self.symbols.items()
                   if symbol_id in identifiers and symbol is not None}
        return symbols

    def _get_types_from_typing_lib(self) -> Dict[str, ISymbol]:
        import typing
        from types import FunctionType

        type_symbols: Dict[str, ISymbol] = {}
        all_types: List[str] = typing.__all__

        for t_id in all_types:
            attr = getattr(typing, t_id)
            if not isinstance(attr, FunctionType):
                type_id: str = t_id if t_id in Type.all_types() else t_id.lower()
                if type_id in Type.all_types():
                    type_symbols[t_id] = Type.all_types()[type_id]

        return type_symbols

    def _get_interop_symbols(self, package: str) -> Dict[str, ISymbol]:
        return Builtin.interop_symbols(package)

    def _get_boa3_builtin_symbols(self) -> Dict[str, ISymbol]:
        return Builtin.boa_symbols()

    def _get_boa3_builtin_package(self, packages: List[str]) -> Dict[str, ISymbol]:
        if len(packages) > 0:
            if len(packages) == 1:
                return Builtin.package_symbols(packages[0])

            if packages[0] == 'interop':
                return self._get_interop_symbols(packages[1])

        return self._get_boa3_builtin_symbols()



from enum import Enum
from typing import Dict, List, Optional, Tuple, Union

from boa3.model.builtin.builtincallable import IBuiltinCallable
from boa3.model.builtin.classmethod.appendmethod import AppendMethod
from boa3.model.builtin.classmethod.clearmethod import ClearMethod
from boa3.model.builtin.classmethod.extendmethod import ExtendMethod
from boa3.model.builtin.classmethod.mapkeysmethod import MapKeysMethod
from boa3.model.builtin.classmethod.mapvaluesmethod import MapValuesMethod
from boa3.model.builtin.classmethod.reversemethod import ReverseMethod
from boa3.model.builtin.classmethod.tobytesmethod import ToBytes as ToBytesMethod
from boa3.model.builtin.classmethod.tointmethod import ToInt as ToIntMethod
from boa3.model.builtin.classmethod.tostrmethod import ToStr as ToStrMethod
from boa3.model.builtin.contract.nep5transferevent import Nep5TransferEvent
from boa3.model.builtin.decorator.metadatadecorator import MetadataDecorator
from boa3.model.builtin.decorator.publicdecorator import PublicDecorator
from boa3.model.builtin.interop.interop import Interop
from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.builtin.method.bytearraymethod import ByteArrayMethod
from boa3.model.builtin.method.createeventmethod import CreateEventMethod, EventType
from boa3.model.builtin.method.lenmethod import LenMethod
from boa3.model.builtin.method.toscripthashmethod import ScriptHashMethod
from boa3.model.builtin.neometadatatype import MetadataTypeSingleton as NeoMetadataType
from boa3.model.callable import Callable
from boa3.model.identifiedsymbol import IdentifiedSymbol
from boa3.model.type.itype import IType


class BoaPackage(str, Enum):
    Contract = 'contract'


class Builtin:
    @classmethod
    def get_symbol(cls, symbol_id: str) -> Optional[Callable]:
        for name, method in vars(cls).items():
            if isinstance(method, IBuiltinCallable) and method.identifier == symbol_id:
                return method

    @classmethod
    def get_by_self(cls, symbol_id: str, self_type: IType) -> Optional[Callable]:
        for name, method in vars(cls).items():
            if (isinstance(method, IBuiltinMethod)
                    and method.identifier == symbol_id
                    and method.validate_self(self_type)):
                return method

    # builtin method
    Len = LenMethod()
    ScriptHash = ScriptHashMethod()
    NewEvent = CreateEventMethod()

    # python builtin class constructor
    ByteArray = ByteArrayMethod()

    # python class method
    SequenceAppend = AppendMethod()
    SequenceClear = ClearMethod()
    SequenceExtend = ExtendMethod()
    SequenceReverse = ReverseMethod()
    DictKeys = MapKeysMethod()
    DictValues = MapValuesMethod()

    # custom class methods
    ConvertToBytes = ToBytesMethod
    ConvertToInt = ToIntMethod
    ConvertToStr = ToStrMethod

    _python_builtins: List[IdentifiedSymbol] = [Len,
                                                ScriptHash,
                                                ByteArray,
                                                SequenceAppend,
                                                SequenceClear,
                                                SequenceExtend,
                                                SequenceReverse,
                                                DictKeys,
                                                DictValues,
                                                ConvertToBytes,
                                                ConvertToInt,
                                                ConvertToStr
                                                ]

    @classmethod
    def interop_symbols(cls, package: str = None) -> Dict[str, IdentifiedSymbol]:
        return {method.identifier: method for method in Interop.interop_symbols(package)}

    # builtin decorator
    Public = PublicDecorator()
    Metadata = MetadataDecorator()

    # boa builtin type
    Event = EventType

    # boa events
    Nep5Transfer = Nep5TransferEvent()

    _boa_builtins: List[IdentifiedSymbol] = [Public,
                                             NewEvent,
                                             Event,
                                             Metadata,
                                             NeoMetadataType
                                             ]

    metadata_fields: Dict[str, Union[type, Tuple[type]]] = \
        {
            'author': (str, type(None)),
            'email': (str, type(None)),
            'description': (str, type(None)),
            'has_storage': bool,
            'is_payable': bool
        }

    @classmethod
    def boa_symbols(cls) -> Dict[str, IdentifiedSymbol]:
        return {symbol.identifier: symbol for symbol in cls._boa_builtins}

    @classmethod
    def package_symbols(cls, package: str = None) -> Dict[str, IdentifiedSymbol]:
        if package in BoaPackage.__members__.values():
            return {symbol.identifier: symbol for symbol in cls._boa_symbols[package]}

        return cls.boa_symbols()

    _boa_symbols: Dict[BoaPackage, List[IdentifiedSymbol]] = {
        BoaPackage.Contract: [Nep5Transfer
                              ]
    }



from boa3.boa3 import Boa3
from boa3.exception.CompilerError import MismatchedTypes, UnfilledArgument
from boa3.model.builtin.interop.interop import Interop
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest


class TestEvent(BoaTest):

    def test_event_without_arguments(self):
        event_name = String('Event').to_bytes(min_length=1)
        expected_output = (
            Opcode.NEWARRAY0    # Main()
            + Opcode.PUSHDATA1      # event()
            + Integer(len(event_name)).to_byte_array(min_length=1)
            + event_name
            + Opcode.SYSCALL
            + Interop.Notify.interop_method_hash
            + Opcode.PUSHNULL       # return
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/event_test/EventWithoutArguments.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_event_with_arguments(self):
        event_name = String('Event').to_bytes(min_length=1)
        expected_output = (
            Opcode.PUSH10        # Main()
            + Opcode.PUSH1          # event(10)
            + Opcode.PACK
            + Opcode.PUSHDATA1
            + Integer(len(event_name)).to_byte_array(min_length=1)
            + event_name
            + Opcode.SYSCALL
            + Interop.Notify.interop_method_hash
            + Opcode.PUSHNULL       # return
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/event_test/EventWithArgument.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_event_with_name(self):
        event_name = String('example').to_bytes(min_length=1)
        expected_output = (
            Opcode.PUSH10       # Main()
            + Opcode.PUSH1
            + Opcode.PACK
            + Opcode.PUSHDATA1      # event(10)
            + Integer(len(event_name)).to_byte_array(min_length=1)
            + event_name
            + Opcode.SYSCALL
            + Interop.Notify.interop_method_hash
            + Opcode.PUSHNULL       # return
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/event_test/EventWithName.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_event_nep5_transfer(self):
        event_name = String('transfer').to_bytes(min_length=1)
        expected_output = (
            Opcode.INITSLOT     # Main()
            + b'\x00\x03'
            + Opcode.LDARG2         # event(from_addr, to_addr, amount)
            + Opcode.LDARG1
            + Opcode.LDARG0
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.PUSHDATA1
            + Integer(len(event_name)).to_byte_array(min_length=1)
            + event_name
            + Opcode.SYSCALL
            + Interop.Notify.interop_method_hash
            + Opcode.PUSHNULL       # return
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/event_test/EventNep5Transfer.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_event_with_return(self):
        path = '%s/boa3_test/test_sc/event_test/EventWithoutTypes.py' % self.dirname
        self.assertCompilerLogs(UnfilledArgument, path)

    def test_event_call_mismatched_type(self):
        path = '%s/boa3_test/test_sc/event_test/MismatchedTypeCallEvent.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)



from __future__ import annotations

from boa3.neo3.core import serialization, utils, IClonable


class StorageItem(serialization.ISerializable, IClonable):
    def __init__(self, value: bytes = None):
        self.value = value if value else b''

    def __len__(self):
        return utils.get_var_size(self.value)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return self.value == other.value

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        writer.write_var_bytes(self.value)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        self.value = reader.read_var_bytes()

    def clone(self) -> StorageItem:
        return StorageItem(self.value)

    def from_replica(self, replica: StorageItem) -> None:
        self.value = replica.value



from typing import Any, Set

from boa3.model.type.collection.sequence.mutable.mutablesequencetype import MutableSequenceType
from boa3.model.type.itype import IType


class ListType(MutableSequenceType):
    """
    A class used to represent Python list type
    """

    def __init__(self, values_type: Set[IType] = None):
        identifier = 'list'
        values_type = self.filter_types(values_type)
        super().__init__(identifier, values_type)

    @property
    def default_value(self) -> Any:
        return list()

    def is_valid_key(self, key_type: IType) -> bool:
        return key_type == self.valid_key

    @property
    def valid_key(self) -> IType:
        from boa3.model.type.type import Type
        return Type.int

    @classmethod
    def build(cls, value: Any):
        if cls._is_type_of(value):
            values_types: Set[IType] = cls.get_types(value)
            return cls(values_types)

    @classmethod
    def _is_type_of(cls, value: Any):
        return type(value) in [list, ListType]

    def __eq__(self, other) -> bool:
        if type(self) != type(other):
            return False
        return self.value_type == other.value_type

    def __hash__(self):
        return hash(self.identifier)



import logging
import os

from boa3.analyser.analyser import Analyser
from boa3.compiler.codegenerator import CodeGenerator
from boa3.compiler.filegenerator import FileGenerator
from boa3.exception.NotLoadedException import NotLoadedException


class Compiler:
    """
    The main compiler class.

    :ivar bytecode: the compiled file as a byte array. Empty by default.
    """

    def __init__(self):
        self.bytecode: bytearray = bytearray()
        self._analyser: Analyser = None

    def compile(self, path: str) -> bytes:
        """
        Load a Python file and tries to compile it

        :param path: the path of the Python file to compile
        :return: the bytecode of the compiled .nef file
        """
        fullpath = os.path.realpath(path)
        filepath, filename = os.path.split(fullpath)

        logging.info('Started compiling\t{0}'.format(filename))
        self._analyse(path)
        return self._compile()

    def compile_and_save(self, path: str, output_path: str):
        """
        Save the compiled file and the metadata files

        :param path: the path of the Python file to compile
        :param output_path: the path to save the generated files
        """
        self.bytecode = self.compile(path)
        self._save(output_path)

    def _analyse(self, path: str):
        """
        Load a Python file and analyses its syntax

        :param path: the path of the Python file to compile
        """
        self._analyser = Analyser.analyse(path, log=True)

    def _compile(self) -> bytes:
        """
        Compile the analysed Python file.

        :return: the compiled file as a bytecode.
        :raise NotLoadedException: raised if none file were analysed
        """
        if not self._analyser.is_analysed:
            raise NotLoadedException
        return CodeGenerator.generate_code(self._analyser)

    def _save(self, output_path: str):
        """
        Save the compiled file and the metadata files

        :param output_path: the path to save the generated files
        :raise NotLoadedException: raised if no file were compiled
        """
        if (self._analyser is None
                or not self._analyser.is_analysed
                or len(self.bytecode) == 0):
            raise NotLoadedException

        generator = FileGenerator(self.bytecode, self._analyser)
        with open(output_path, 'wb+') as nef_file:
            nef_bytes = generator.generate_nef_file()
            nef_file.write(nef_bytes)
            nef_file.close()

        with open(output_path.replace('.nef', '.manifest.json'), 'wb+') as manifest_file:
            manifest_bytes = generator.generate_manifest_file()
            manifest_file.write(manifest_bytes)
            manifest_file.close()

        from zipfile import ZipFile, ZIP_DEFLATED
        with ZipFile(output_path.replace('.nef', '.nefdbgnfo'), 'w', ZIP_DEFLATED) as nef_debug_info:
            debug_bytes = generator.generate_nefdbgnfo_file()
            nef_debug_info.writestr(os.path.basename(output_path.replace('.nef', '.debug.json')), debug_bytes)
